package com.randomsoftware.vncclient;

import danger.ui.*;
import danger.app.*;
import danger.util.*;
import java.util.*;

public class CellManager
{
   static final int MAX_LIVE_PIXEL_COUNT = 800 * 600;

   public CellManager()
   {
      _usedCells = new CellList();
      _freeCells = new CellList();
      _zoomCells = new CellList();
   }

   public void releaseMemory()
   {
      // Un-grid all in-use cells
      Cell cell;
      while ((cell = _usedCells.popFirst()) != null) {
         cell.removeFromGrid();
      }

      // Throw away all our cells, to free memory
      _usedCells = new CellList();
      _freeCells = new CellList();
      _zoomCells = new CellList();
   }

   public void releaseMyCells (FrameBuffer buf)
   {
      Cell next;
      for (Cell walk = _usedCells.getFirst(); walk != null; walk = next) {
         next = _usedCells.getNext(walk);
         if (walk._buf == buf) {
            walk.removeFromGrid();
            _usedCells.remove(walk);
            _freeCells.insertTail(walk);
         }
      }

      // If we have more cells allocated than we want, give some
      // up for garbage collection
      while ((_usedCells.size() + _freeCells.size() > getMaxLiveCellCount()) && (!_freeCells.empty())) {
         _freeCells.popFirst();
      }
   }

   public void releaseZoomCache()
   {
      for (Cell walk = _zoomCells.getFirst(); walk != null; walk = _zoomCells.getNext(walk)) {
         Cell liveCell = walk._zoomCell;
         if (liveCell != null) {
            liveCell._zoomCell = null;
            walk._zoomCell = null;
            walk._ixCell = -1;
            walk._iyCell = -1;
         }
      }
      _zoomCells = new CellList();
   }

   public Cell allocateCell (FrameBuffer buf, int ixCell, int iyCell)
   {
      Cell cell = null;
      if (_usedCells.size() + _freeCells.size() < getMaxLiveCellCount()) {
         cell = new Cell();
      }
      if (cell == null && !_freeCells.empty()) {
         cell = _freeCells.popFirst();
      }
      if (cell == null && !_usedCells.empty()) {
         int best = 0;
         for (Cell walk = _usedCells.getFirst(); walk != null; walk = _usedCells.getNext(walk)) {
            int score = walk.getScore(buf, ixCell, iyCell);
            if (score > best) {
               best = score;
               cell = walk;
            }
         }
         if (cell != null) {
            _usedCells.remove(cell);
            cell.removeFromGrid();
         }
      }
      if (cell == null) {
         cell = new Cell();
      }
      if (cell != null) {
         cell.addToGrid (buf, ixCell, iyCell);
         _usedCells.insertTail(cell);
      }
      return cell;
   }

   public Cell allocateZoomCell (int zoomPercent)
   {
      Cell cell;
      if (_zoomCells.size() < getMaxZoomCellCount(zoomPercent)) {
         cell = new Cell (zoomPercent);
      } else {
         cell = _zoomCells.popFirst();
         if (cell._zoomCell != null) {
            // Clear the live cell's back pointer to this zoom cell
            cell._zoomCell._zoomCell = null;
            cell._ixCell = -1;
            cell._iyCell = -1;
            // And forget about the old live cell.
            cell._zoomCell = null;
         }
      }
      _zoomCells.insertTail(cell);
      return cell;
   }

   static int getMaxLiveCellCount()
   {
      int cPixelsCell = Cell.WIDTH * Cell.HEIGHT;
      return (MAX_LIVE_PIXEL_COUNT + cPixelsCell-1) / cPixelsCell;
   }

   static int getMaxZoomCellCount(int zoomPercent)
   {
      // We want enough cells to cover the screen, even if we're
      // rendering at an unaligned location (which we will be).
      int cxZoomCell = Cell.WIDTH * zoomPercent / 100;
      int cyZoomCell = Cell.HEIGHT * zoomPercent / 100;
      return ((Compat.SCREEN_HEIGHT + cyZoomCell - 1) / cyZoomCell + 1) *
             ((Compat.SCREEN_WIDTH + cxZoomCell - 1) / cxZoomCell + 1);
   }

   CellList _usedCells;
   CellList _freeCells;
   CellList _zoomCells;
}

class Cell
{
   static final int WIDTH = 256;
   static final int HEIGHT = 128;

   public Cell _prev = null;
   public Cell _next = null;
   public Cell _zoomCell = null;

   public Bitmap bmp;
   public int _cxSize;
   public int _cySize;

   public int _ixCell;
   public int _iyCell;
   public FrameBuffer _buf;

   public Cell()
   {
      _cxSize = WIDTH;
      _cySize = HEIGHT;
      bmp = new Bitmap (_cxSize, _cySize);
   }

   public Cell (int zoomPercent)
   {
      _cxSize = WIDTH * zoomPercent / 100;
      _cySize = HEIGHT * zoomPercent / 100;
      _ixCell = -2;
      _iyCell = -2;
      bmp = new Bitmap (_cxSize, _cySize);
   }

   public void addToGrid (FrameBuffer buf, int ixCell, int iyCell)
   {
      _buf = buf;
      _ixCell = ixCell;
      _iyCell = iyCell;
      _buf._cellMap[_iyCell][_ixCell] = this;
      _buf.floodCell(this);
   }

   public void removeFromGrid()
   {
      releaseZoomCache();
      _buf._cellMap[_iyCell][_ixCell] = null;
      _buf._unknownRegions.add (_ixCell * WIDTH, _iyCell * HEIGHT, WIDTH, HEIGHT);
      _buf._mustRecalcUpdatingSize = true;
   }

   public Cell getZoomCache(int zoomPercent)
   {
      if (_zoomCell == null) {
         Cell got = VncClient.getCellManager().allocateZoomCell(zoomPercent);
         _zoomCell = got;
         _zoomCell._ixCell = _ixCell;
         _zoomCell._iyCell = _iyCell;
         got._zoomCell = this;
         got.bmp.copyScaled (bmp);
      }
      return _zoomCell;
   }

   // Someone has updated the content in this cell.  If we have cached
   // a ZoomCell for it, throw away that link.
   public void releaseZoomCache()
   {
      Cell kill = _zoomCell;
      if (kill != null) {
         // Clear the zoom cell's back pointer here.
         kill._zoomCell = null;
         kill._ixCell = -1;
         kill._iyCell = -1;
         // And forget we ever had a zoom cell.
         _zoomCell = null;
         // Then move the zoom cell to the front, for quick recycling later.
         VncClient.getCellManager()._zoomCells.remove(kill);
         VncClient.getCellManager()._zoomCells.insertHead(kill);
      }
   }

   // The further this cell is from the specified coordinates,
   // the higher the "score" we give back.  This helps the manager
   // decide which cell to recycle when it needs a new one--it
   // prefers to grab cells from other sessions, or if it has to
   // use one in this session it wants to grab cells from far away.
   public int getScore (FrameBuffer buf, int ixCell, int iyCell)
   {
      int score = 0;
      if (buf != _buf) {
         score += WIDTH * HEIGHT;
      }
      if (buf != _buf || !buf.isCellVisible (_ixCell, _iyCell)) {
         score += Math.abs(ixCell - _ixCell);
         score += Math.abs(iyCell - _iyCell);
      }
      return score;
   }
}

class CellList
{
   public Cell getFirst()
   {
      return _first;
   }

   public Cell getNext (Cell cell)
   {
      return cell._next;
   }

   public Cell popFirst()
   {
      Cell cell = _first;
      if (cell != null) {
         remove (cell);
      }
      return cell;
   }

   public void insertHead (Cell cell)
   {
      cell._next = _first;
      cell._prev = null;
      if (_first != null)
         _first._prev = cell;
      else
         _last = cell;
      _first = cell;
      _size++;
   }

   public void insertTail (Cell cell)
   {
      cell._prev = _last;
      cell._next = null;
      if (_last != null)
         _last._next = cell;
      else
         _first = cell;
      _last = cell;
      _size++;
   }

   public void remove (Cell cell)
   {
      if (cell._prev == null)
         _first = cell._next;
      else
         cell._prev._next = cell._next;
      if (cell._next == null)
         _last = cell._prev;
      else
         cell._next._prev = cell._prev;
      _size--;
   }

   public boolean empty()
   {
      return (_size == 0);
   }

   public int size()
   {
      return _size;
   }

   protected Cell _first = null;
   protected Cell _last = null;
   protected int _size = 0;
}

