package com.randomsoftware.vncclient;

import danger.app.*;
import danger.ui.*;
import danger.util.*;
import java.util.*;

public class DataStream
{
   public DataStream()
   {
      _buffers = new DataBufferList();
      _totalBytes = 0;
   }

   public synchronized int avail()
   {
      return _totalBytes;
   }

   public synchronized int peekInt1()
   {
      DataBuffer head = _buffers.getFirst();
      return unsignedByteToInt(head.data[ head.ibFirst ]);
   }

   public synchronized int readByte()
   {
      DataBuffer head = _buffers.getFirst();
      byte rc = head.data[ head.ibFirst ];
      skip(1);
      return rc;
   }

   public synchronized int readInt1()
   {
      DataBuffer head = _buffers.getFirst();
      int rc = unsignedByteToInt(head.data[ head.ibFirst ]);
      skip(1);
      return rc;
   }

   public synchronized int readInt2()
   {
      int rc;
      DataBuffer head = _buffers.getFirst();
      if (head.cbTotal >= 2) {
         rc = (unsignedByteToInt(head.data[ head.ibFirst ]) << 8) +
               unsignedByteToInt(head.data[ head.ibFirst+1 ]);
         skip(2);
      } else {
         rc = (readInt1() << 8) + readInt1();
      }
      return rc;
   }

   public synchronized int readInt4()
   {
      int rc;
      DataBuffer head = _buffers.getFirst();
      if (head.cbTotal >= 4) {
         rc = (unsignedByteToInt(head.data[ head.ibFirst ]) << 24) +
              (unsignedByteToInt(head.data[ head.ibFirst+1 ]) << 16) +
              (unsignedByteToInt(head.data[ head.ibFirst+2 ]) << 8) +
               unsignedByteToInt(head.data[ head.ibFirst+3 ]);
         skip(4);
      } else {
         rc = (readInt1() << 24) + (readInt1() << 16) + (readInt1() << 8) + readInt1();
      }
      return rc;
   }

   public synchronized void readBytes (byte[] out, int offset, int count)
   {
      while (count != 0) {
         DataBuffer head = _buffers.getFirst();
         int cbCopy = (head.cbTotal < count) ? head.cbTotal : count;
         System.arraycopy (head.data, head.ibFirst, out, offset, cbCopy);
         offset += cbCopy;
         count -= cbCopy;
         skip (cbCopy);
      }
   }

   public synchronized void skip (int count)
   {
      while (count > 0) {
         DataBuffer head = _buffers.getFirst();
         int cbSkip = (head.cbTotal < count) ? head.cbTotal : count;
         head.ibFirst += cbSkip;
         head.cbTotal -= cbSkip;
         _totalBytes -= cbSkip;
         count -= cbSkip;
         if (head.cbTotal == 0) {
            _buffers.remove(head);
            releaseBuffer (head);
         }
      }
   }

   public synchronized void skipAll()
   {
      while (!_buffers.empty()) {
         DataBuffer head = _buffers.getFirst();
         skip (head.cbTotal);
      }
   }

   public synchronized void append (DataBuffer buf)
   {
      _buffers.insertTail (buf);
      _totalBytes += buf.cbTotal;
   }

   public synchronized DataBuffer allocateBuffer()
   {
      DataBuffer buf;
      if ((buf = getFreeList().popFirst()) != null) {
         buf.clear();
      } else {
         buf = new DataBuffer();
      }
      return buf;
   }

   public synchronized void releaseBuffer (DataBuffer buf)
   {
      if (getFreeList().size() < FREELIST_MAX_SIZE) {
         getFreeList().insertTail(buf);
      }
   }

   public static DataBufferList getFreeList()
   {
      return _freeList;
   }

   public static void releaseMemory()
   {
      // Throw away anything on the freelist
      _freeList = new DataBufferList();
   }

   public static int unsignedByteToInt (byte bb)
   {
      return ((int)bb) & 0xFF;
   }

   static final int FREELIST_MAX_SIZE = 5;

   static DataBufferList _freeList = new DataBufferList();
   DataBufferList _buffers;
   int _totalBytes;
}


class DataBufferList
{
   public DataBuffer getFirst()
   {
      return _first;
   }

   public DataBuffer getNext (DataBuffer db)
   {
      return db._next;
   }

   public DataBuffer popFirst()
   {
      DataBuffer db = _first;
      if (db != null) {
         remove (db);
      }
      return db;
   }

   public void insertTail (DataBuffer db)
   {
      db._prev = _last;
      db._next = null;
      if (_last != null)
         _last._next = db;
      else
         _first = db;
      _last = db;
      _size++;
   }

   public void remove (DataBuffer db)
   {
      if (db._prev == null)
         _first = db._next;
      else
         db._prev._next = db._next;
      if (db._next == null)
         _last = db._prev;
      else
         db._next._prev = db._prev;
      _size--;
   }

   public boolean empty()
   {
      return (_size == 0);
   }

   public int size()
   {
      return _size;
   }

   protected DataBuffer _first = null;
   protected DataBuffer _last = null;
   protected int _size = 0;
}


class DataBuffer
{
   byte[] data;
   int ibFirst;
   int cbTotal;

   public DataBuffer()
   {
      this(DEFAULT_SIZE);
   }

   public DataBuffer (int size)
   {
      data = new byte[size];
      ibFirst = 0;
      cbTotal = 0;
   }

   public void clear()
   {
      ibFirst = 0;
      cbTotal = 0;
   }

   public void writeInt1 (int val)
   {
      data[ cbTotal++ ] = (byte)val;
   }

   public void writeInt2 (int val)
   {
      data[ cbTotal++ ] = (byte)((val >>  8) & 0xFF);
      data[ cbTotal++ ] = (byte)(val & 0xFF);
   }

   public void writeInt4 (int val)
   {
      data[ cbTotal++ ] = (byte)((val >> 24) & 0xFF);
      data[ cbTotal++ ] = (byte)((val >> 16) & 0xFF);
      data[ cbTotal++ ] = (byte)((val >>  8) & 0xFF);
      data[ cbTotal++ ] = (byte)(val & 0xFF);
   }

   public void writeBytes (byte[] in, int offset, int count)
   {
      System.arraycopy (in, offset, data, cbTotal, count);
      cbTotal += count;
   }

   protected char hexchar (byte b)
   {
      if (b < 10) {
         return (char)('0'+b);
      } else {
         return (char)('A'+b-10);
      }
   }

   public String toHexString()
   {
      StringBuffer out = new StringBuffer();

      for (int ii = 0; ii < cbTotal; ++ii) {
         out.append( (ii==0?"0x":" ") + 
                     hexchar( (byte)((DataStream.unsignedByteToInt(data[ibFirst+ii]) >> 4) & 0xF) ) +
                     hexchar( (byte)(DataStream.unsignedByteToInt(data[ibFirst+ii]) & 0xF) ));
      }

      return out.toString();
   }

   public String toLimitedHexString (int limit)
   {
      StringBuffer out = new StringBuffer();

      if (limit > cbTotal) {
         limit = cbTotal;
      }

      for (int ii = 0; ii < limit; ++ii) {
         out.append( (ii==0?"0x":" ") + 
                     hexchar( (byte)((DataStream.unsignedByteToInt(data[ibFirst+ii]) >> 4) & 0xF) ) +
                     hexchar( (byte)(DataStream.unsignedByteToInt(data[ibFirst+ii]) & 0xF) ));
      }

      return out.toString();
   }

   DataBuffer _prev = null;
   DataBuffer _next = null;
   static final int DEFAULT_SIZE = 8192;
}

