package com.randomsoftware.vncclient;

import danger.app.*;
import danger.ui.*;
import danger.util.*;
import java.util.*;

// The FrameBuffer class maps out a sparse grid of Cell objects, each
// representing a portion of the remote desktop's display space.  The
// class also has a seperate Region object that indicates where we have
// data and where we don't; it's important to note that, just because
// we have a cell covering a spot, doesn't mean we have data for it.

// The FrameBuffer is responsible for rendering the remote desktop
// onto your display.  To do this it works with the RemoteView object,
// which knows what part of the screen you're viewing, and at what
// zoom level, and where your cursor is.  That job includes gelling
// over areas where the display is known to be out-of-date.

public class FrameBuffer
{
   public FrameBuffer (Session session)
   {
      _session = session;
      _view = _session.getRemoteView();
      _unknownRegions = new Region();
      _updatingRegions = new Region();
      _tempRegions = new Region();
      _cellMapWidth = 0;
      _cellMapHeight = 0;
      _cellMap = null;
      _rCell = new Rect();
      _rDesktop = new Rect();
      _mustRecalcUpdatingSize = true;

      if (_bmpUnknownBlend == null) {
         _bmpUnknownBlend = new Bitmap (BITMAP_UNKNOWN_WIDTH, BITMAP_UNKNOWN_HEIGHT);
         Pen pp = _bmpUnknownBlend.createPen();
         pp.setColor (_clrUnknown);
         pp.fillRect (0, 0, BITMAP_UNKNOWN_WIDTH, BITMAP_UNKNOWN_HEIGHT);
      }
   }

   public void destroy()
   {
      VncClient.getCellManager().releaseMyCells (this);
   }

   public void rebuild()
   {
      VncClient.getCellManager().releaseMyCells (this);

      _cellMapWidth = (_view.cxRemoteDesktop + Cell.WIDTH-1) / Cell.WIDTH;
      _cellMapHeight = (_view.cyRemoteDesktop + Cell.HEIGHT-1) / Cell.HEIGHT;
      _cellMap = new Cell[_cellMapHeight][_cellMapWidth];

      _unknownRegions.clear();
      _unknownRegions.add(0, 0, _view.cxRemoteDesktop, _view.cyRemoteDesktop);
      _updatingRegions.clear();
      _mustRecalcUpdatingSize = true;

      refreshVisible (REFRESH_IF_UNKNOWN);
   }


   // Determines which cells (by IX) are drawn on the left-edge and right-edge,
   // and which (by IY) are drawn on the top-edge and bottom-edge.

   public int getLeftCell()
   {
      int ixbScreenLeft = _view.ixbScreenCenter - _view.getVisibleBufferWidth()/2;
      int ixCell = ixbScreenLeft / Cell.WIDTH;
      return (ixCell < 0) ? 0 : (ixCell >= _cellMapWidth) ? (_cellMapWidth-1) : ixCell;
   }

   public int getRightCell()
   {
      int ixbScreenRight = _view.ixbScreenCenter + _view.getVisibleBufferWidth()/2;
      int ixCell = ixbScreenRight / Cell.WIDTH;
      return (ixCell < 0) ? 0 : (ixCell >= _cellMapWidth) ? (_cellMapWidth-1) : ixCell;
   }

   public int getTopCell()
   {
      int iybScreenTop = _view.iybScreenCenter - _view.getVisibleBufferHeight()/2;
      int iyCell = iybScreenTop / Cell.HEIGHT;
      return (iyCell < 0) ? 0 : (iyCell >= _cellMapHeight) ? (_cellMapHeight-1) : iyCell;
   }

   public int getBottomCell()
   {
      int iybScreenBottom = _view.iybScreenCenter + _view.getVisibleBufferHeight()/2;
      int iyCell = iybScreenBottom / Cell.HEIGHT;
      return (iyCell < 0) ? 0 : (iyCell >= _cellMapHeight) ? (_cellMapHeight-1) : iyCell;
   }


   // Determines if a particular cell is on-screen now, and where

   public boolean isCellVisible (int ix, int iy)
   {
      return (ix >= getLeftCell() && ix <= getRightCell() &&
              iy >= getTopCell() && iy <= getBottomCell());
   }

   public void getCellRectOnBuffer (Rect out, int ix, int iy)
   {
      out.left = ix * Cell.WIDTH;
      out.right = out.left + Cell.WIDTH;
      out.top = iy * Cell.HEIGHT;
      out.bottom = out.top + Cell.HEIGHT;

      if (out.right > _view.cxRemoteDesktop) {
         out.right = _view.cxRemoteDesktop;
      }
      if (out.bottom > _view.cyRemoteDesktop) {
         out.bottom = _view.cyRemoteDesktop;
      }
   }

   public void getCellRectOnScreen (Rect out, int ix, int iy)
   {
      getCellRectOnBuffer (out, ix, iy);
      _view.scaleBufferRectToScreen (out);
   }


   // Ensures that the currently-visible region is up-to-date, starting
   // calls to the remote desktop for a refresh if necessary

   static final int REFRESH_IF_UNKNOWN = 1;
   static final int DISCARD_AND_REFRESH = 2;

   public void refreshVisible (int mode)
   {
      _view.getVisibleBuffer (_rDesktop);
      _mustRecalcUpdatingSize = true;

      if (mode == DISCARD_AND_REFRESH) {
         _unknownRegions.add (_rDesktop.left, _rDesktop.top, _rDesktop.right-_rDesktop.left, _rDesktop.bottom-_rDesktop.top);
         _updatingRegions.subtract (_rDesktop.left, _rDesktop.top, _rDesktop.right-_rDesktop.left, _rDesktop.bottom-_rDesktop.top);
      }

      for (RegionRect walk = _unknownRegions.walkAffected(_rDesktop,_rCell); walk != null; walk = _unknownRegions.walkAffectedNext(_rDesktop,_rCell,walk)) {
         _tempRegions.clear();
         _tempRegions.add(_rCell.left, _rCell.top, _rCell.right-_rCell.left, _rCell.bottom-_rCell.top);

         for (RegionRect busy = _updatingRegions.getFirst(); busy != null; busy = _updatingRegions.getNext(busy)) {
            _tempRegions.subtract(busy.left, busy.top, busy.right-busy.left, busy.bottom-busy.top);
         }

         for (RegionRect need = _tempRegions.getFirst(); need != null; need = _tempRegions.getNext(need)) {
            Protocol.requestUpdate (_session.getConnection(), need.left, need.top, need.right-need.left, need.bottom-need.top, Protocol.FULL_UPDATE);
            _updatingRegions.add (need.left, need.top, need.right-need.left, need.bottom-need.top);
         }
      }

      RemoteScreen screen = _session.getMyRemoteScreen();
      if (screen != null) {
         screen.updateRefreshOverlay();
      }
   }


   // Find out how much data we're currently asking for

   public int getUpdatingSize()
   {
      if (_mustRecalcUpdatingSize) {
         int countNow = 0;
         for (RegionRect walk = _updatingRegions.getFirst(); walk != null; walk = _updatingRegions.getNext(walk)) {
            countNow += (walk.right-walk.left) * (walk.bottom-walk.top);
         }

         // If we're in the middle of decoding an image, we might have
         // partially processed some portion that we're not prepared
         // to update yet--and yet we want to track incremental progress.
         // Find out how much we'll have still-updating after that image
         // is done, and interpolate between those Before and After counts
         // based on the decoder's current progress.

         // TODO: Note that this precise counter is disabled right now!
         // Turns out that all this region manipulation is horribly CPU
         // intensive, and we're doing it all the time during updates.
         // Turning it off isn't too bad since the refresh overlay still
         // updates nicely (not entirely sure why, and don't want to
         // think about it too hard in case it breaks).  In order to
         // turn this back on, we'll need some kind of throttle here:
         // if it's been less than a quarter-second since the last full
         // recalc, then just start a timer to wake us up later and try
         // again then.

         int countFinal = countNow;

         ImageDecoder decoder = _session.getImageDecoder();
         if (false && 
             decoder.isActive() &&
             decoder.getPercentComplete() != 0 &&
             decoder.getPercentComplete() != 100) {
            _tempRegions.clear();
            for (RegionRect walk = _updatingRegions.getFirst(); walk != null; walk = _updatingRegions.getNext(walk)) {
               _tempRegions.add (walk.left, walk.top, walk.right-walk.left, walk.bottom-walk.top);
            }
            _tempRegions.subtract (decoder._xx, decoder._yy, decoder._cx, decoder._cy);

            int countSoon = 0;
            for (RegionRect walk = _tempRegions.getFirst(); walk != null; walk = _tempRegions.getNext(walk)) {
               countSoon += (walk.right-walk.left) * (walk.bottom-walk.top);
            }

            countFinal = countNow + (countSoon - countNow) * decoder.getPercentComplete() / 100;
         }

         _lastUpdatingSize = countFinal;
         _mustRecalcUpdatingSize = false;
      }

      return _lastUpdatingSize;
   }


   // The CellManager has just allocated a cell; flood its bitmap with
   // the unknown-contents color.

   public void floodCell (Cell cell)
   {
      Pen pp = cell.bmp.createPen();
      pp.setColor (_clrUnknown);
      pp.fillRect (0, 0, Cell.WIDTH, Cell.HEIGHT);
   }



   // Hey, we got some content from the server!  Paint it into cells,
   // and update our various regions.

   public void update (Bitmap bmp, int xx, int yy, int cx, int cy, int xFrom, int yFrom)
   {
      int iCellLeft = xx / Cell.WIDTH;
      int iCellRight = (xx+cx) / Cell.WIDTH;
      int iCellTop = yy / Cell.HEIGHT;
      int iCellBottom = (yy+cy) / Cell.HEIGHT;

      if (iCellLeft >= _cellMapWidth || iCellTop >= _cellMapHeight) {
         return;
      }
      if (iCellRight >= _cellMapWidth) {
         iCellRight = _cellMapWidth-1;
      }
      if (iCellBottom >= _cellMapHeight) {
         iCellBottom = _cellMapHeight-1;
      }

      _updatingRegions.subtract (xx, yy, cx, cy);
      _unknownRegions.subtract (xx, yy, cx, cy);
      _mustRecalcUpdatingSize = true;

      RemoteScreen screen = _session.getMyRemoteScreen();

      for (int iy = iCellTop; iy <= iCellBottom; ++iy) {
         for (int ix = iCellLeft; ix <= iCellRight; ++ix) {
            if (_cellMap[iy][ix] == null) {
               VncClient.getCellManager().allocateCell (this, ix, iy);
            }

            if (_cellMap[iy][ix] != null) {
               int xBuffer = ix * Cell.WIDTH + ((ix == iCellLeft) ? (xx % Cell.WIDTH) : 0);
               int yBuffer = iy * Cell.HEIGHT + ((iy == iCellTop) ? (yy % Cell.HEIGHT) : 0);
               int cxBuffer = Cell.WIDTH;
               int cyBuffer = Cell.HEIGHT;

               if (((xBuffer + cxBuffer - 1) / Cell.WIDTH) >= ix) {
                  cxBuffer = (ix + 1) * Cell.WIDTH - xBuffer;
               }
               if (xBuffer + cxBuffer > xx + cx) {
                  cxBuffer = xx + cx - xBuffer;
               }

               if (((yBuffer + cyBuffer - 1) / Cell.HEIGHT) >= iy) {
                  cyBuffer = (iy + 1) * Cell.HEIGHT - yBuffer;
               }
               if (yBuffer + cyBuffer > yy + cy) {
                  cyBuffer = yy + cy - yBuffer;
               }

               _rCell.left = xBuffer % Cell.WIDTH;
               _rCell.right = _rCell.left + cxBuffer;
               _rCell.top = yBuffer % Cell.HEIGHT;
               _rCell.bottom = _rCell.top + cyBuffer;

               _rDesktop.left = xBuffer - xx + xFrom;
               _rDesktop.top = yBuffer - yy + yFrom;
               _rDesktop.right = _rDesktop.left + cxBuffer;
               _rDesktop.bottom = _rDesktop.top + cyBuffer;

               Bitmap.copyBits (_cellMap[iy][ix].bmp, _rCell, bmp, _rDesktop);
               _cellMap[iy][ix].releaseZoomCache();

               if (screen != null) {
                  getCellRectOnScreen (_rCell, ix, iy);
                  screen.invalidate (_rCell.left, _rCell.top, _rCell.right, _rCell.bottom);
               }
            }
         }
      }

      if (screen != null) {
         screen.updateRefreshOverlay();
      }
   }


   // Read content from one part of the frame buffer, and write it to
   // another.  This is one of the RFB encoding methods.

   public void copyContents (int xTo, int yTo, int cxCopy, int cyCopy, int xFrom, int yFrom)
   {
      int iFromLeft = xFrom / Cell.WIDTH;
      int iFromRight = (xFrom+cxCopy) / Cell.WIDTH;
      int iFromTop = yFrom / Cell.HEIGHT;
      int iFromBottom = (yFrom+cyCopy) / Cell.HEIGHT;
      _mustRecalcUpdatingSize = true;

      for (int iy = iFromTop; iy <= iFromBottom; ++iy) {
         for (int ix = iFromLeft; ix <= iFromRight; ++ix) {
            if (_cellMap[iy][ix] != null) {

               int xBuffer = ix * Cell.WIDTH + ((ix == iFromLeft) ? (xFrom % Cell.WIDTH) : 0);
               int yBuffer = iy * Cell.HEIGHT + ((iy == iFromTop) ? (yFrom % Cell.HEIGHT) : 0);
               int cxBuffer = Cell.WIDTH;
               int cyBuffer = Cell.HEIGHT;

               if (((xBuffer + cxBuffer - 1) / Cell.WIDTH) >= ix) {
                  cxBuffer = (ix + 1) * Cell.WIDTH - xBuffer;
               }
               if (xBuffer + cxBuffer > xFrom + cxCopy) {
                  cxBuffer = xFrom + cxCopy - xBuffer;
               }

               if (((yBuffer + cyBuffer - 1) / Cell.HEIGHT) >= iy) {
                  cyBuffer = (iy + 1) * Cell.HEIGHT - yBuffer;
               }
               if (yBuffer + cyBuffer > yFrom + cyCopy) {
                  cyBuffer = yFrom + cyCopy - yBuffer;
               }

               int xTarget = xBuffer + xTo - xFrom;
               int yTarget = yBuffer + yTo - yFrom;

               update (_cellMap[iy][ix].bmp,
                        xTarget,
                        yTarget,
                        cxBuffer,
                        cyBuffer,
                        xBuffer % Cell.WIDTH,
                        yBuffer % Cell.HEIGHT);
            }
         }
      }
   }


   // Flood a rectangle with the specified color (which should be
   // already encoded in the correct on-screen format).  Does NOT
   // mark the specified region as being finished, since there may
   // still be work to do here to get it to look right.

   public void flood (SubRectangle[] rects, int xMin, int yMin, int xMax, int yMax, int count)
   {
      int iCellLeft = xMin / Cell.WIDTH;
      int iCellRight = xMax / Cell.WIDTH;
      int iCellTop = yMin / Cell.HEIGHT;
      int iCellBottom = yMax / Cell.HEIGHT;

      if (iCellLeft >= _cellMapWidth || iCellTop >= _cellMapHeight) {
         return;
      }
      if (iCellRight >= _cellMapWidth) {
         iCellRight = _cellMapWidth-1;
      }
      if (iCellBottom >= _cellMapHeight) {
         iCellBottom = _cellMapHeight-1;
      }

      RemoteScreen screen = _session.getMyRemoteScreen();

      for (int iy = iCellTop; iy <= iCellBottom; ++iy) {
         for (int ix = iCellLeft; ix <= iCellRight; ++ix) {
            if (_cellMap[iy][ix] == null) {
               VncClient.getCellManager().allocateCell (this, ix, iy);
            }
            if (_cellMap[iy][ix] != null) {
               int xCellMin = ix * Cell.WIDTH;
               int yCellMin = iy * Cell.HEIGHT;
               int xCellMax = xCellMin + Cell.WIDTH;
               int yCellMax = yCellMin + Cell.HEIGHT;

               byte[] pixels = _cellMap[iy][ix].bmp.getPixels();
               int pitch = _cellMap[iy][ix].bmp.getStride();
               _cellMap[iy][ix].releaseZoomCache();

               for (int ii = 0; ii < count; ++ii) {
                  if ((rects[ii].cx == 1) && (rects[ii].cy == 1)) {

                     // We'll frequently get a request to update exactly one pixel.
                     // That's easy to do.
                     if ((rects[ii].xx >= xCellMin) && (rects[ii].xx < xCellMax) &&
                         (rects[ii].yy >= yCellMin) && (rects[ii].yy < yCellMax)) {
                        int offset = (rects[ii].yy % Cell.HEIGHT) * pitch + (rects[ii].xx % Cell.WIDTH) * 2;
                        pixels[offset] = (byte)((rects[ii].color>>8) & 0xFF);
                        pixels[offset+1] = (byte)((rects[ii].color) & 0xFF);
                     }

                  } else {

                     // Otherwise, we'll use fillRect to update many pixels at once.
                     if ((rects[ii].xx < xCellMax) && (rects[ii].xx+rects[ii].cx > xCellMin) &&
                         (rects[ii].yy < yCellMax) && (rects[ii].yy+rects[ii].cy > yCellMin)) {
                        int xFloodMin = rects[ii].xx - xCellMin;
                        int yFloodMin = rects[ii].yy - yCellMin;
                        int xFloodMax = xFloodMin + rects[ii].cx;
                        int yFloodMax = yFloodMin + rects[ii].cy;
                        if (xFloodMin < 0) {
                           xFloodMin = 0;
                        }
                        if (yFloodMin < 0) {
                           yFloodMin = 0;
                        }
                        if (xFloodMax > Cell.WIDTH) {
                           xFloodMax = Cell.WIDTH;
                        }
                        if (yFloodMax > Cell.HEIGHT) {
                           yFloodMax = Cell.HEIGHT;
                        }

                        byte b1 = (byte)((rects[ii].color>>8) & 0xFF);
                        byte b2 = (byte)((rects[ii].color) & 0xFF);

                        for (int yy = yFloodMin; yy < yFloodMax; ++yy) {
                           int offset = yy * pitch + xFloodMin * 2;
                           for (int xx = xFloodMin; xx < xFloodMax; ++xx) {
                              pixels[offset++] = b1;
                              pixels[offset++] = b2;
                           }
                        }
                     }

                  }
               }
            }
         }
      }

      _mustRecalcUpdatingSize = true;
   }


   // Some methods of using the frame buffer don't automatically mark
   // it up-to-date.  When that happens, the decoder will eventually
   // want to tell us when it has finished with a region.

   public void finished (int xx, int yy, int cx, int cy)
   {
      _updatingRegions.subtract (xx, yy, cx, cy);
      _unknownRegions.subtract (xx, yy, cx, cy);
      _mustRecalcUpdatingSize = true;

      RemoteScreen screen = _session.getMyRemoteScreen();
      if (screen != null) {
         screen.updateRefreshOverlay();
         screen.invalidate();
      }
   }


   // The big Paint method.  Renders what we can, then draws unknown
   // areas with a moving image.  Also draws the local cursor.  And
   // does it all zoomed.  Remember that all this work is done by
   // consulting this Session's RemoteView object (which tells us the
   // zoom level, etc).

   public void paint (Pen pp)
   {
      // What region of the screen needs to be updated?
      Rect rDisplay = pp.getClip();

      if (_cellMap == null) {
         pp.setColor(_clrUnknown);
         pp.fillRect(rDisplay);
         return;
      }

      // If we're zoomed out so far that the framebuffer doesn't fill
      // the display, first wash everything in a basic background
      // color so we won't get artifacts around the edges.
      if (_view.cxRemoteDesktop < _view.getLocalWidth() || _view.cyRemoteDesktop < _view.getLocalHeight()) {
         pp.setColor(128, 128, 128);
         pp.fillRect(rDisplay);
      }

      // Find out what cells are being rendered
      int ixLeft = getLeftCell();
      int ixRight = getRightCell();
      int iyTop = getTopCell();
      int iyBottom = getBottomCell();
      _view.getVisibleBuffer (_rDesktop);

      // Draw each cell in turn
      for (int iy = iyTop; iy <= iyBottom; ++iy) {
         for (int ix = ixLeft; ix <= ixRight; ++ix) {
            getCellRectOnScreen (_rCell, ix, iy);

            // If we don't have a cell here, draw it in blue; otherwise,
            // draw the cell.  Remember that cells are flooded with the
            // unknown color, so if any cell isn't fully populated it's
            // irrelevent.
            if (_cellMap[iy][ix] == null) {
               pp.setColor(_clrUnknown);
               pp.fillRect(_rCell);
            } else if (_view.zoomPercent == 100) {
               pp.drawBitmap (_rCell.left, _rCell.top, _cellMap[iy][ix].bmp, 0, 0, _rCell.right-_rCell.left, _rCell.bottom-_rCell.top);
            } else {
               Cell zoom = _cellMap[iy][ix].getZoomCache(_view.zoomPercent);
               pp.drawBitmap (_rCell.left, _rCell.top, zoom.bmp, 0, 0, _rCell.right-_rCell.left, _rCell.bottom-_rCell.top);
            }
         }
      }

      // Then draw a gel over the unknown areas, so the user
      // knows that they're, well, unknown.  Blending over a
      // fully-unknown region is a no-op visually, so don't
      // worry about that.
      for (RegionRect walk = _unknownRegions.walkAffected(_rDesktop,_rCell); walk != null; walk = _unknownRegions.walkAffectedNext(_rDesktop,_rCell,walk)) {
         _view.scaleBufferRectToScreen (_rCell);
         for (int yy = _rCell.top; yy < _rCell.bottom; yy += BITMAP_UNKNOWN_HEIGHT) {
            for (int xx = _rCell.left; xx < _rCell.right; xx += BITMAP_UNKNOWN_WIDTH) {
               int cx = _rCell.right - xx;
               int cy = _rCell.bottom - yy;
               if (cx >= BITMAP_UNKNOWN_WIDTH && cy >= BITMAP_UNKNOWN_HEIGHT) {
                  Compat.Pen_drawBitmap (pp, xx, yy, _bmpUnknownBlend, BITMAP_UNKNOWN_BLEND);
               } else {
                  Pen p2 = pp.clonePen();
                  p2.setClip (xx, yy, xx+cx, yy+cy);
                  Compat.Pen_drawBitmap (p2, xx, yy, _bmpUnknownBlend, BITMAP_UNKNOWN_BLEND);
               }
            }
         }
      }
   }


   Session _session;
   RemoteView _view;
   Cell[][] _cellMap;
   int _cellMapWidth;
   int _cellMapHeight;
   int _lastUpdatingSize;
   boolean _mustRecalcUpdatingSize;
   Region _unknownRegions;
   Region _updatingRegions;
   Region _tempRegions;
   Rect _rCell;
   Rect _rDesktop;
   static Color _clrUnknown = new Color (0, 70, 160);
   static Bitmap _bmpUnknownBlend = null;
   static final int BITMAP_UNKNOWN_WIDTH = 32;
   static final int BITMAP_UNKNOWN_HEIGHT = 32;
   static final int BITMAP_UNKNOWN_BLEND = 128;
}

