package com.randomsoftware.vncclient;

import danger.app.*;
import danger.ui.*;
import danger.util.*;
import java.util.*;

class ImageDecoder
{
   public ImageDecoder()
   {
      _isActive = false;
      _scratch = null;
      _percentComplete = 0;
      _rre_subRects = null;
      _lastYieldWasVoluntary = false;
   }

   public boolean isActive()
   {
      return _isActive;
   }

   public int getPercentComplete()
   {
      return _percentComplete;
   }

   public void reset()
   {
      _isActive = false;
      _percentComplete = 0;
   }

   public boolean lastYieldWasVoluntary()
   {
      return _lastYieldWasVoluntary;
   }

   public boolean start (Session session, DataStream stream)
   {
      _lastYieldWasVoluntary = false;

      // Can't start decoding unless there's a 4-byte header.
      if (stream.avail() < 4) {
         return false;
      }

      // This a frame-update request, right?
      stream.skip(2);
      _unfinishedCount = stream.readInt2();
      if (_unfinishedCount == 0)
         return false;

      _isActive = true;
      _needHeader = true;
      _percentComplete = 0;
      return true;
   }

   public void decode (Session session, FrameBuffer frameBuffer, DataStream stream)
   {
      _timeYield = System.currentTimeMillis() + YIELD_RATE_MILLISECONDS;
      _lastYieldWasVoluntary = false;

      while (_isActive) {
         if (_needHeader) {
            if (stream.avail() < 12) {
               return;
            }
            _xx = stream.readInt2();
            _yy = stream.readInt2();
            _cx = stream.readInt2();
            _cy = stream.readInt2();
            _encoding = stream.readInt4();
            _needHeader = false;
            _initialized = false;
         }

         switch (_encoding) {
            case ENCODING_RAW:
               if (!decodeRaw (session, frameBuffer, stream)) {
                  return;
               }
               break;

            case ENCODING_COPYRECT:
               if (!decodeCopyRect (session, frameBuffer, stream)) {
                  return;
               }
               break;

            case ENCODING_RRE:
            case ENCODING_CORRE:
               if (!decodeRRE (session, frameBuffer, stream)) {
                  return;
               }
               break;

            case ENCODING_HEXTILE:
               if (!decodeHextile (session, frameBuffer, stream)) {
                  return;
               }
               break;

            case ENCODING_PSEUDO_CURSOR:
               if (!decodeCursor (session, frameBuffer, stream)) {
                  return;
               }
               break;

            default:
               stream.skipAll();
Debug.out("can't handle unknown image encoding "+_encoding);
               _isActive = false;
               return;
         }

         _unfinishedCount--;
         if (_unfinishedCount == 0) {
            _percentComplete = 100;
            _isActive = false;
         } else {
            _percentComplete = 0;
            _needHeader = true;
         }

         if (System.currentTimeMillis() > _timeYield) {
            _lastYieldWasVoluntary = true;
            session.sendLowPriorityEvent (new Event (SessionConnectionEvent.RECEIVED));
         }
      }
   }

   public boolean decodeCursor (Session session, FrameBuffer frameBuffer, DataStream stream)
   {
      // Wait until the whole cursor is ready to be processed

      int cbRequired;
      cbRequired = _cy * _cx * 2;	     // the cursor's image data
      cbRequired += _cy * ((_cx + 7) / 8);   // the cursor's mask data
      if (stream.avail() < cbRequired) {
         return false;
      }

      // Then prepare the cursor image

      SessionCursor cursor = session.getCursor();
      Bitmap bmpCursor = cursor.createBitmap (_xx, _yy, _cx, _cy);
      byte[] pixels = bmpCursor.getPixels();
      int cbPitchCursor = bmpCursor.getStride();
      int cbPitchIncoming = _cx * 2;

      for (int yy = 0; yy < _cy; ++yy) {
         stream.readBytes (pixels, yy * cbPitchCursor, cbPitchIncoming);
      }

      int clrTransparent = (0x1F << 11) | (0x3F << 5) | (0);
      int xTransparent = _cx-1;
      int yTransparent = _cy-1;
      for (int yy = 0; yy < _cy; ++yy) {
         int xx = 0;
         for (int ii = 0; ii < (_cx+7)/8; ++ii) {
            int mask = stream.readInt1();
            for (int bit = 0; xx < _cx && bit < 8; ++xx,++bit) {
               if ((mask & (1 << (7-bit))) == 0) {
                  pixels[ yy * cbPitchCursor + xx * 2 ]    = (byte)(clrTransparent >> 8);
                  pixels[ yy * cbPitchCursor + xx * 2 +1 ] = (byte)(clrTransparent);
                  xTransparent = xx;
                  yTransparent = yy;
               }
            }
         }
      }

      int color = bmpCursor.getPoint (xTransparent, yTransparent);
      int rr = (color >> 19) & 0x1F;
      int gg = (color >> 10) & 0x3F;
      int bb = (color >> 3) & 0x1F;
      bmpCursor.setTransparentColor ((bb << 11) | (gg << 5) | rr);

      cursor.setConfig (SessionCursor.CONFIG_AUTOMATIC);
      return true;
   }

   public boolean decodeRaw (Session session, FrameBuffer frameBuffer, DataStream stream)
   {
      // If we're just starting, keep track of how much there is to go

      if (!_initialized) {
         _initialized = true;
         _raw_scanlinesLeft = _cy;
         resizeScratch();
      }

      // We'll crack the available data until we finish, run out or give up

      while (_raw_scanlinesLeft != 0) {

         // Is there enough data available to justify working forward now,
         // or should we wait?

         int scanlinesAvail = stream.avail() / (_cx * 2);
         if ((scanlinesAvail < _raw_scanlinesLeft) && (scanlinesAvail < _scratch.getHeight())) {
            return false;
         }

         // Okay, looks like we can make progress here.  How many scanlines
         // can we process on this iteration?  That's the lessor of
         // how many lines are left to process, how much data is available,
         // and how tall our scratch bitmap is (it's always wide enough).

         int linesToFill = _raw_scanlinesLeft;
         if (linesToFill > _scratch.getHeight()) {
            linesToFill = _scratch.getHeight();
         }
         if (linesToFill > scanlinesAvail) {
            linesToFill = scanlinesAvail;
         }

         byte[] scratchData = _scratch.getPixels();
         int scratchPitch = _scratch.getStride();
         int inboundPitch = _cx * 2;

         for (int yy = 0; yy < linesToFill; ++yy) {
            stream.readBytes (scratchData, yy*scratchPitch, inboundPitch);
         }

         // Then draw the scratch bitmap onto the Frame Buffer

         frameBuffer.update (_scratch, _xx, _yy + _cy - _raw_scanlinesLeft, _cx, linesToFill, 0, 0);

         // And move on to the next region

         _raw_scanlinesLeft -= linesToFill;
      }

      // Must be done!
      return true;
   }

   public boolean decodeCopyRect (Session session, FrameBuffer frameBuffer, DataStream stream)
   {
      if (stream.avail() < 4) {
         return false;
      }
      int xFrom = stream.readInt2();
      int yFrom = stream.readInt2();

      frameBuffer.copyContents (_xx, _yy, _cx, _cy, xFrom, yFrom);

      return true;
   }

   public boolean decodeRRE (Session session, FrameBuffer frameBuffer, DataStream stream)
   {
      int recordSize = (_encoding == ENCODING_CORRE) ? 6 : 10;

      // If we haven't already, we need to read the header off this thing;
      // that tells us how many subrectangles to expect.

      if (!_initialized) {
         if (stream.avail() < 6) {
            return false;
         }

         _rre_subRectsTotal = stream.readInt4();
         _rre_subRectsLeft = _rre_subRectsTotal;

         resizeSubrects (1);
         _rre_subRects[0].xx = (short)_xx;
         _rre_subRects[0].yy = (short)_yy;
         _rre_subRects[0].cx = (short)_cx;
         _rre_subRects[0].cy = (short)_cy;
         _rre_subRects[0].color = (short)stream.readInt2();
         frameBuffer.flood (_rre_subRects, _xx, _yy, _xx+_cx, _yy+_cy, 1);
         _initialized = true;
      }

      // Now process as many subrectangles as we can.

      while (_rre_subRectsLeft > 0) {

         // Find out how many we can process; the more the better, but
         // we can't process partial rects and we don't want to do more
         // than N at a time (in case we need to yield).

         int recordCount = stream.avail() / recordSize;
         if (recordCount == 0) {
            break;
         }
         if (recordCount > _rre_subRectsLeft) {
            recordCount = _rre_subRectsLeft;
         }
         if (recordCount > RRE_MAX_SUBRECTS_AT_ONCE) {
            recordCount = RRE_MAX_SUBRECTS_AT_ONCE;
         }
         resizeSubrects (recordCount);

         // Fill in our array with subrect data

         int xMin = 0;
         int yMin = 0;
         int xMax = 0;
         int yMax = 0;
         for (int ii = 0; ii < recordCount; ++ii) {
            if (_encoding == ENCODING_CORRE) {
               _rre_subRects[ii].color = (short)stream.readInt2();
               _rre_subRects[ii].xx = (short)(stream.readInt1() + _xx);
               _rre_subRects[ii].yy = (short)(stream.readInt1() + _yy);
               _rre_subRects[ii].cx = (short)stream.readInt1();
               _rre_subRects[ii].cy = (short)stream.readInt1();
            } else {
               _rre_subRects[ii].color = (short)stream.readInt2();
               _rre_subRects[ii].xx = (short)(stream.readInt2() + _xx);
               _rre_subRects[ii].yy = (short)(stream.readInt2() + _yy);
               _rre_subRects[ii].cx = (short)stream.readInt2();
               _rre_subRects[ii].cy = (short)stream.readInt2();
            }

            if (ii == 0) {
               xMin = _rre_subRects[ii].xx;
               yMin = _rre_subRects[ii].yy;
               xMax = _rre_subRects[ii].xx+_rre_subRects[ii].cx;
               yMax = _rre_subRects[ii].yy+_rre_subRects[ii].cy;
            } else {
               if (xMin > _rre_subRects[ii].xx)
                  xMin = _rre_subRects[ii].xx;
               if (yMin > _rre_subRects[ii].yy)
                  yMin = _rre_subRects[ii].yy;
               if (xMax < _rre_subRects[ii].xx+_rre_subRects[ii].cx)
                  xMax = _rre_subRects[ii].xx+_rre_subRects[ii].cx;
               if (yMax < _rre_subRects[ii].yy+_rre_subRects[ii].cy)
                  yMax = _rre_subRects[ii].yy+_rre_subRects[ii].cy;
            }
         }

         // And update the screen.  This will take a long time,
         // so when we're done we want to yield back to let other
         // things run for a moment.  We'll also put a fake you-just-
         // got-data event on the queue, so it will come back here and
         // do more in a second.

         frameBuffer.flood (_rre_subRects, xMin, yMin, xMax, yMax, recordCount);
         _rre_subRectsLeft -= recordCount;

         if (System.currentTimeMillis() > _timeYield) {
            _lastYieldWasVoluntary = true;
            session.sendLowPriorityEvent (new Event (SessionConnectionEvent.RECEIVED));
            break;
         }
      }

      // And now we're done; tell the frame buffer that it's up-to-date

      if (_rre_subRectsTotal == 0) {
         _percentComplete = 100;
      } else {
         _percentComplete = 100 * (_rre_subRectsTotal - _rre_subRectsLeft) / _rre_subRectsTotal;
      }

      if (_rre_subRectsLeft == 0) {
         frameBuffer.finished (_xx, _yy, _cx, _cy);
      }

      return (_rre_subRectsLeft == 0);
   }

   public boolean decodeHextile (Session session, FrameBuffer frameBuffer, DataStream stream)
   {
      if (!_initialized) {
         _initialized = true;
         _hex_subTilesAcross = ((_cx + HEXTILE_SIZE-1) / HEXTILE_SIZE);
         _hex_subTilesDown = ((_cy + HEXTILE_SIZE-1) / HEXTILE_SIZE);
         _hex_subTilesTotal = _hex_subTilesAcross * _hex_subTilesDown;
         _hex_subTilesLeft = _hex_subTilesTotal;
         _hex_tileCode = HEXTILE_CODE_UNKNOWN;
         _hex_colorBack = (short)0x0000;
         _hex_colorFore = (short)0xFFFF;
         _hex_needTileHeader = true;
         resizeScratch (HEXTILE_SIZE, HEXTILE_SIZE);
      }

      byte[] scratchData = _scratch.getPixels();
      int scratchPitch = _scratch.getStride();

      while (_hex_subTilesLeft > 0) {

         int iTile = _hex_subTilesTotal - _hex_subTilesLeft;
         int ixTile = iTile % _hex_subTilesAcross;
         int iyTile = iTile / _hex_subTilesAcross;
         int xBufferTile = _xx + ixTile * HEXTILE_SIZE;
         int yBufferTile = _yy + iyTile * HEXTILE_SIZE;
         int cxBufferTile = (ixTile+1 == _hex_subTilesAcross) ? ((_cx-1) % HEXTILE_SIZE + 1) : HEXTILE_SIZE;
         int cyBufferTile = (iyTile+1 == _hex_subTilesDown) ? ((_cy-1) % HEXTILE_SIZE + 1) : HEXTILE_SIZE;

         // Each tile starts with a header byte

         if (_hex_tileCode == HEXTILE_CODE_UNKNOWN) {
            if (stream.avail() < 1) {
               return false;
            }
            _hex_tileCode = stream.readInt1();
         }

         // Different tiles have different encoding methods

         if ((_hex_tileCode & HEXTILE_CODE_RAW) != 0) {

            // Raw tiles are just full of 16x16 pixel data.  That's easy.
            if (stream.avail() < cxBufferTile * cyBufferTile * 2) {
               return false;
            }

            int sourcePitch = cxBufferTile * 2;
            for (int yy = 0; yy < cyBufferTile; ++yy) {
               stream.readBytes (scratchData, yy*scratchPitch, sourcePitch);
            }

         } else {

            // Non-raw tiles have a variable header composed of different fields.
            // One of the most important things we'll get here is the count-of-subrects.
            if (_hex_needTileHeader) {
               int bytesNeeded = 0;
               if ((_hex_tileCode & HEXTILE_CODE_BACK) != 0)
                  bytesNeeded += 2;
               if ((_hex_tileCode & HEXTILE_CODE_FORE) != 0)
                  bytesNeeded += 2;
               if ((_hex_tileCode & HEXTILE_CODE_SUBRECTS) != 0)
                  bytesNeeded += 1;

               if (stream.avail() < bytesNeeded) {
                  return false;
               }

               if ((_hex_tileCode & HEXTILE_CODE_BACK) != 0)
                  _hex_colorBack = (short)stream.readInt2();
               if ((_hex_tileCode & HEXTILE_CODE_FORE) != 0)
                  _hex_colorFore = (short)stream.readInt2();
               if ((_hex_tileCode & HEXTILE_CODE_SUBRECTS) != 0)
                  _hex_subRectsTotal = stream.readInt1();
               else
                  _hex_subRectsTotal = 0;

               _hex_needTileHeader = false;
            }

            // Make sure we have all the subrect data on tap before we proceed
            if (_hex_subRectsTotal != 0) {
               int bytesNeeded = _hex_subRectsTotal;
               if ((_hex_tileCode & HEXTILE_CODE_COLORED) != 0)
                  bytesNeeded *= 4;
               else
                  bytesNeeded *= 2;
               if (stream.avail() < bytesNeeded) {
                  return false;
               }
            }

            // Flood the tile with the background color
            for (int yy = 0; yy < cyBufferTile; ++yy) {
               int offset = yy * scratchPitch;
               for (int xx = 0; xx < cxBufferTile; ++xx) {
                  scratchData[offset++] = (byte)(_hex_colorBack >> 8);
                  scratchData[offset++] = (byte)(_hex_colorBack);
               }
            }

            // Draw any subrects
            for (int ii = 0; ii < _hex_subRectsTotal; ++ii) {
               short color = ((_hex_tileCode & HEXTILE_CODE_COLORED) != 0) ? (short)stream.readInt2() : _hex_colorFore;
               int xy = stream.readInt1();
               int wh = stream.readInt1();

               int x1 = (xy>>4) & 0xF;
               int y1 = xy & 0xF;
               int x2 = x1 + ((wh>>4) & 0xF) + 1;
               int y2 = y1 + (wh & 0xF) + 1;

               for (int yy = y1; yy < y2; ++yy) {
                  int offset = yy * scratchPitch + x1 * 2;
                  for (int xx = x1; xx < x2; ++xx) {
                     scratchData[offset++] = (byte)(color >> 8);
                     scratchData[offset++] = (byte)(color);
                  }
               }
            }
         }

         // Now  we've populated the scratch image; push it to the frame buffer

         frameBuffer.update (_scratch, xBufferTile, yBufferTile, cxBufferTile, cyBufferTile, 0, 0);

         // Time to move on to the next tile

         _hex_tileCode = HEXTILE_CODE_UNKNOWN;
         _hex_needTileHeader = true;
         _hex_subTilesLeft--;

         if (System.currentTimeMillis() > _timeYield) {
            _lastYieldWasVoluntary = true;
            session.sendLowPriorityEvent (new Event (SessionConnectionEvent.RECEIVED));
            break;
         }
      }

      return (_hex_subTilesLeft == 0);
   }

   public void resizeScratch()
   {
      if (_scratch == null || _scratch.getWidth() < _cx) {
         int height = _cy;
         if (_cx * _cy * 2 > SCRATCH_BITMAP_MAX_SIZE) {
            height = SCRATCH_BITMAP_MAX_SIZE / (_cx * 2);
         }
         _scratch = new Bitmap (_cx, height);
      }
   }

   public void resizeScratch (int cx, int cy)
   {
      if (_scratch == null || _scratch.getWidth() < cx || _scratch.getHeight() < cy) {
         _scratch = new Bitmap (cx, cy);
      }
   }

   public void resizeSubrects (int wanted)
   {
      if ((_rre_subRects == null) || (wanted > _rre_subRects.length)) {
         _rre_subRects = new SubRectangle[wanted];
         for (int ii = 0; ii < wanted; ++ii) {
            _rre_subRects[ii] = new SubRectangle();
         }
      }
   }

   int _msgType;
   boolean _isActive;
   boolean _needHeader;
   boolean _initialized;
   int _unfinishedCount;
   int _xx;
   int _yy;
   int _cx;
   int _cy;
   int _encoding;
   int _percentComplete;
   long _timeYield;
   boolean _lastYieldWasVoluntary;

   int _raw_scanlinesLeft;
   int _rre_subRectsLeft;
   int _rre_subRectsTotal;
   SubRectangle[] _rre_subRects;
   boolean _hex_needTileHeader;
   int _hex_subTilesLeft;
   int _hex_subTilesTotal;
   int _hex_subTilesAcross;
   int _hex_subTilesDown;
   int _hex_subRectsTotal;
   int _hex_tileCode;
   short _hex_colorBack;
   short _hex_colorFore;

   Bitmap _scratch;
   public static final int SCRATCH_BITMAP_MAX_SIZE = 65535;
   public static final int YIELD_RATE_MILLISECONDS = 100;

   public static final int RRE_MAX_SUBRECTS_AT_ONCE = 250;

   public static final int HEXTILE_SIZE = 16;
   public static final int HEXTILE_CODE_UNKNOWN = -1;
   public static final int HEXTILE_CODE_RAW = 0x01;
   public static final int HEXTILE_CODE_BACK = 0x02;
   public static final int HEXTILE_CODE_FORE = 0x04;
   public static final int HEXTILE_CODE_SUBRECTS = 0x08;
   public static final int HEXTILE_CODE_COLORED = 0x10;
   public static final int HEXTILE_NEED_SUBRECTS = -1;

   public static final int ENCODING_RAW = 0;
   public static final int ENCODING_COPYRECT = 1;
   public static final int ENCODING_RRE = 2;
   public static final int ENCODING_CORRE = 4;
   public static final int ENCODING_HEXTILE = 5;
   public static final int ENCODING_ZLIB = 6;
   public static final int ENCODING_TIGHT = 7;
   public static final int ENCODING_ZLIBHEX = 8;
   public static final int ENCODING_ZRLE = 16;
   public static final int ENCODING_PSEUDO_CURSOR = -239;
   public static final int ENCODING_PSEUDO_DESKTOP = -223;
}

class SubRectangle
{
   short xx;
   short yy;
   short cx;
   short cy;
   short color;
}

