package com.randomsoftware.vncclient;

import danger.app.*;
import danger.ui.*;
import danger.util.*;
import danger.crypto.*;
import danger.audio.*;
import java.util.*;

// The Protocol class encapsulates most of the VNC wire protocol.
// The interface is thread-safe, but should not be called on the
// foreground thread since many of the APIs block behind socket sends.

public class Protocol
{
   // Inbound Messages
   static public final int MESSAGEID_IN_BUFFERUPDATE = 0;
   static public final int MESSAGEID_IN_BELL = 2;
   static public final int MESSAGEID_IN_SERVER_CLIP = 3;

   // Outbound Messages
   static public final int MESSAGEID_OUT_SETPIXELFORMAT = 0;
   static public final int MESSAGEID_OUT_SETENCODINGS = 2;
   static public final int MESSAGEID_OUT_UPDATEREQUEST = 3;
   static public final int MESSAGEID_OUT_KEYEVENT = 4;
   static public final int MESSAGEID_OUT_POINTEREVENT = 5;
   static public final int MESSAGEID_OUT_CLIENT_CLIP = 6;

   // Messages
   static final int FULL_UPDATE = 1;
   static final int INCREMENTAL_UPDATE = 2;
   static public void requestUpdate (SessionConnection conn, int xx, int yy, int cx, int cy, int type)
   {
      SendThread out = conn.getSendThread();
      DataBuffer buf = out.begin();

      Debug.out("requesting "+(type == FULL_UPDATE ? "full" : "incremental")+" update over "+xx+","+yy+" "+cx+"x"+cy);

      buf.writeInt1(MESSAGEID_OUT_UPDATEREQUEST);
      buf.writeInt1((type == INCREMENTAL_UPDATE) ? 1 : 0);
      buf.writeInt2(xx);
      buf.writeInt2(yy);
      buf.writeInt2(cx);
      buf.writeInt2(cy);

      out.end(buf);
   }

   static public void sendPointerEvent (SessionConnection conn, int xx, int yy, int buttonMask)
   {
      SendThread out = conn.getSendThread();
      DataBuffer buf = out.begin();

      buf.writeInt1(MESSAGEID_OUT_POINTEREVENT);
      buf.writeInt1(buttonMask);
      buf.writeInt2(xx);
      buf.writeInt2(yy);

      out.end(buf);
   }

   static public void sendPointerDoubleClick (SessionConnection conn, int xx, int yy, int maskNow, int iButton)
   {
      int maskButton = buttonIndexToMask(iButton);
      int maskUp = maskNow & (~maskButton);
      int maskDown = maskNow | maskButton;

      SendThread out = conn.getSendThread();
      DataBuffer buf = out.begin();

      buf.writeInt1(MESSAGEID_OUT_POINTEREVENT);
      buf.writeInt1(maskDown);
      buf.writeInt2(xx);
      buf.writeInt2(yy);

      buf.writeInt1(MESSAGEID_OUT_POINTEREVENT);
      buf.writeInt1(maskUp);
      buf.writeInt2(xx);
      buf.writeInt2(yy);

      buf.writeInt1(MESSAGEID_OUT_POINTEREVENT);
      buf.writeInt1(maskDown);
      buf.writeInt2(xx);
      buf.writeInt2(yy);

      buf.writeInt1(MESSAGEID_OUT_POINTEREVENT);
      buf.writeInt1(maskUp);
      buf.writeInt2(xx);
      buf.writeInt2(yy);

      out.end(buf);
   }

   static public void sendKeyEvent (SessionConnection conn, int keysym, boolean down)
   {
      SendThread out = conn.getSendThread();
      DataBuffer buf = out.begin();

      buf.writeInt1(MESSAGEID_OUT_KEYEVENT);
      buf.writeInt1((down) ? 1 : 0);  // up/down
      buf.writeInt2(0);               // padding
      buf.writeInt4(keysym);          // keysym

      out.end(buf);
   }

   static public void sendKeySequence (SessionConnection conn, int[] keys, int count)
   {
      SendThread out = conn.getSendThread();

      // Every key requires 16 bytes to send (wow!).  We might have to break
      // the transmission into pieces to send all these keys, depending
      // on how many entries are in the array and how big our send-buffer is.
      // DON'T use this method for shift,alt,control etc--just printables.

      for (int ii = 0; ii < count; ) {
         DataBuffer buf = out.begin();
         int keysPerLoop = buf.data.length / 16;

         int thisLoop = count - ii;
         if (thisLoop > keysPerLoop)
            thisLoop = keysPerLoop;

         for (int jj = 0; jj < thisLoop; ++jj,++ii) {
            buf.writeInt1(MESSAGEID_OUT_KEYEVENT);
            buf.writeInt1(1);  // down
            buf.writeInt2(0);  // padding
            buf.writeInt4(keys[ii]);  // keysym

            buf.writeInt1(MESSAGEID_OUT_KEYEVENT);
            buf.writeInt1(0);  // up
            buf.writeInt2(0);  // padding
            buf.writeInt4(keys[ii]);  // keysym
         }

         out.end(buf);
      }
   }

   static public void sendClipboard (SessionConnection conn)
   {
      if (conn.getState() == SessionState.CONNECTED) {
         String cut = Pasteboard.getString();
         if (cut != null) {
            byte[] data = cut.getBytes();
            SendThread out = conn.getSendThread();
            DataBuffer buf = out.begin();

            int length = data.length;
            if (length + 8 > buf.data.length) {
               length = buf.data.length - 8;
            }

            buf.writeInt1(MESSAGEID_OUT_CLIENT_CLIP);
            buf.writeInt1(0);
            buf.writeInt1(0);
            buf.writeInt1(0);  // padding
            buf.writeInt4(length);
            buf.writeBytes(data, 0, length);

            out.end(buf);
         }
      }
   }

   static public void sendInt1 (SessionConnection conn, int val)
   {
      SendThread out = conn.getSendThread();
      DataBuffer buf = out.begin();

      buf.clear();
      buf.writeInt1(val);

      out.end(buf);
   }

   static public void sendPassword (SessionConnection conn, byte[] d1, byte[] d2)
   {
      SendThread out = conn.getSendThread();
      DataBuffer buf = out.begin();

      buf.writeBytes(d1, 0, 8);
      buf.writeBytes(d2, 0, 8);

      out.end(buf);
   }

   static public void sendVersionRequest (SessionConnection conn, int verMajor, int verMinor)
   {
      SendThread out = conn.getSendThread();
      DataBuffer buf = out.begin();

      Debug.out ("asking server to use version "+verMajor+"."+verMinor);

      buf.data[0] = 'R';
      buf.data[1] = 'F';
      buf.data[2] = 'B';
      buf.data[3] = ' ';
      buf.data[4] = (byte)('0' +  verMajor/100);
      buf.data[5] = (byte)('0' + (verMajor%100)/10);
      buf.data[6] = (byte)('0' +  verMajor%10);
      buf.data[7] = '.';
      buf.data[8] = (byte)('0' +  verMinor/100);
      buf.data[9] = (byte)('0' + (verMinor%100)/10);
      buf.data[10] = (byte)('0' + verMinor%10);
      buf.data[11] = '\n';
      buf.ibFirst = 0;
      buf.cbTotal = 12;

      out.end(buf);
   }

   static public void sendClientInit (SessionConnection conn, boolean makeExclusive)
   {
      SendThread out = conn.getSendThread();
      DataBuffer buf = out.begin();

      buf.writeInt1(makeExclusive ? 0 : 1);

      out.end(buf);
   }

   static public void sendSetPixelFormat (SessionConnection conn)
   {
      SendThread out = conn.getSendThread();
      DataBuffer buf = out.begin();

      buf.writeInt1(MESSAGEID_OUT_SETPIXELFORMAT);
      buf.writeInt1(0);   // padding
      buf.writeInt1(0);   // padding
      buf.writeInt1(0);   // padding
      buf.writeInt1(16);  // bits-per-pixel (16-bit encoding)
      buf.writeInt1(16);  // depth (565 == 16)
      buf.writeInt1(0);   // bigendian
      buf.writeInt1(1);   // color info follows
      buf.writeInt2(0x1F);   // red-max
      buf.writeInt2(0x3F);   // green-max
      buf.writeInt2(0x1F);   // blue-max
      buf.writeInt1(0);      // red-shift
      buf.writeInt1(5);      // green-shift
      buf.writeInt1(11);     // blue-shift
      buf.writeInt1(0);   // padding
      buf.writeInt1(0);   // padding
      buf.writeInt1(0);   // padding

      out.end(buf);
   }

   static public void sendSetEncodings (SessionConnection conn)
   {
      SendThread out = conn.getSendThread();
      DataBuffer buf = out.begin();

      buf.writeInt1(MESSAGEID_OUT_SETENCODINGS);
      buf.writeInt1(0);   // padding
      buf.writeInt2(5);   // number of encodings that follow
      buf.writeInt4(ImageDecoder.ENCODING_HEXTILE);
      buf.writeInt4(ImageDecoder.ENCODING_CORRE);
      buf.writeInt4(ImageDecoder.ENCODING_RRE);
      buf.writeInt4(ImageDecoder.ENCODING_RAW);
      buf.writeInt4(ImageDecoder.ENCODING_PSEUDO_CURSOR);

      out.end(buf);
   }

   static public int buttonIndexToMask (int iButton)
   {
      return 1 << iButton;
   }
}

