package com.randomsoftware.vncclient;

import danger.ui.*;
import danger.app.*;
import java.io.*;
import danger.system.*;

// The RemoteView class represents the pseudo-persistent state for
// a session: what part of the screen are you viewing, at what zoom
// level, and where did you leave the cursor, and are you full screen?
// This encapsulation lives in the Session and is only discarded
// when the session itself is destroyed.

public class RemoteView
{
   public int zoomPercent;
   public boolean hideRemoteCursor;
   public boolean showTyping;
   public boolean fullScreen;

   public int ixbScreenCenter;
   public int iybScreenCenter;
   public int ixbCursor;
   public int iybCursor;

   public int cxRemoteDesktop;
   public int cyRemoteDesktop;

   public RemoteView (Session session)
   {
      _session = session;

      ixbScreenCenter = Compat.SCREEN_WIDTH/2;
      iybScreenCenter = Compat.SCREEN_HEIGHT/2;
      ixbCursor = Compat.SCREEN_WIDTH/2;
      iybCursor = Compat.SCREEN_HEIGHT/2;
      cxRemoteDesktop = 0;
      cyRemoteDesktop = 0;

      updateHostSettings (_session.getHostSettings());
   }

   public void updateHostSettings (HostSettings host)
   {
      showTyping = host.showTyping;
      setHideRemoteCursor (host.hideRemoteCursor);
      setZoomPercent (host.zoomPercent);
      setFullScreen (host.fullScreen);
   }

   public void setRemoteSize (int width, int height)
   {
      cxRemoteDesktop = width;
      cyRemoteDesktop = height;
      railViewToDesktop(REFRESH_IF_MOVED);
      getMyFrameBuffer().rebuild();
   }

   public void setFullScreen (boolean bb)
   {
      fullScreen = bb;

      RemoteScreen screen = getMyRemoteScreen();
      if (screen != null) {
         screen.setFullScreen(fullScreen);
      }

      railViewToDesktop(ALWAYS_REFRESH);
   }

   public void setHideRemoteCursor (boolean hide)
   {
      hideRemoteCursor = hide;
      int config = _session.getCursor().getConfig();
      if ((config == SessionCursor.CONFIG_HIDE_REMOTE_CURSOR) && (!hide)) {
         _session.getCursor().setConfig (SessionCursor.CONFIG_SHOW_REMOTE_CURSOR);
      } else if ((config == SessionCursor.CONFIG_SHOW_REMOTE_CURSOR) && (hide)) {
         _session.getCursor().setConfig (SessionCursor.CONFIG_HIDE_REMOTE_CURSOR);
      }
   }

   public void zoomIn()
   {
      if (zoomPercent == 50)
         setZoomPercent(100);
      else if (zoomPercent == 100)
         setZoomPercent(200);
      else
         setZoomPercent(400);
   }

   public void zoomOut()
   {
      if (zoomPercent == 400)
         setZoomPercent(200);
      else if (zoomPercent == 200)
         setZoomPercent(100);
      else
         setZoomPercent(50);
   }

   public void setZoomPercent (int zoom)
   {
      zoomPercent = zoom;
      VncClient.getCellManager().releaseZoomCache();
      railViewToDesktop(ALWAYS_REFRESH);
   }

   public void moveByKeymap (int code)
   {
      int dx = 0;
      int dy = 0;

      if (code == RawKeymap.BUTTON_UP_ARROW) {
         dy = 0 - getVisibleBufferHeight() * 3 / 4;
      } else if (code == RawKeymap.BUTTON_DOWN_ARROW) {
         dy = getVisibleBufferHeight() * 3 / 4;
      } else if (code == RawKeymap.BUTTON_LEFT_ARROW) {
         dx = 0 - getVisibleBufferWidth() * 3 / 4;
      } else if (code == RawKeymap.BUTTON_RIGHT_ARROW) {
         dx = getVisibleBufferWidth() * 3 / 4;
      }

      int ixbScreenCenterOld = ixbScreenCenter;
      int iybScreenCenterOld = iybScreenCenter;

      ixbScreenCenter += dx;
      iybScreenCenter += dy;

      railViewToDesktop(ALWAYS_REFRESH | DONT_MOVE_CURSOR);

      _session.getCursor().moveByPixel (ixbScreenCenter - ixbScreenCenterOld, iybScreenCenter - iybScreenCenterOld);
   }


   // Returns the count-of-local-pixels size of the phone display,
   // taking into account whether we're fullscreen or not

   public int getLocalWidth()
   {
      return Compat.SCREEN_WIDTH;
   }

   public int getLocalHeight()
   {
      if (fullScreen) {
         return Compat.SCREEN_HEIGHT;
      } else {
         return Compat.SCREEN_HEIGHT - ((Compat.SCREEN_HEIGHT <= 160) ? 24 : 32);
      }
   }


   // Returns the count-of-buffer-pixels size of the phone
   // display, taking into account fullscreen and zoom.

   public int getVisibleBufferWidth()
   {
      return getLocalWidth() * 100 / zoomPercent;
   }

   public int getVisibleBufferHeight()
   {
      return getLocalHeight() * 100 / zoomPercent;
   }


   // Determines which pixels on the remote buffer are currently
   // visible on the screen, considering your position and zoom

   public void getVisibleBuffer (Rect out)
   {
      out.left = ixbScreenCenter - getVisibleBufferWidth()/2;
      out.right = ixbScreenCenter + getVisibleBufferWidth()/2;
      out.top = iybScreenCenter - getVisibleBufferHeight()/2;
      out.bottom = iybScreenCenter + getVisibleBufferHeight()/2;

      if (out.left < 0) {
         out.left = 0;
      }
      if (out.right > cxRemoteDesktop) {
         out.right = cxRemoteDesktop;
      }
      if (out.top < 0) {
         out.top = 0;
      }
      if (out.bottom > cyRemoteDesktop) {
         out.bottom = cyRemoteDesktop;
      }
   }


   // Given a rectangle in buffer-coordinates, returns the on-screen
   // coordinates where that rectangle would be drawn.

   public void scaleBufferRectToScreen (Rect rr)
   {
      rr.left = (getLocalWidth()/2) + (rr.left - ixbScreenCenter) * zoomPercent / 100;
      rr.right = (getLocalWidth()/2) + (rr.right - ixbScreenCenter) * zoomPercent / 100;
      rr.top = (getLocalHeight()/2) + (rr.top - iybScreenCenter) * zoomPercent / 100;
      rr.bottom = (getLocalHeight()/2) + (rr.bottom - iybScreenCenter) * zoomPercent / 100;
   }


   // Returns the size of the remote desktop, given your preferred zoom level.

   public int getScaledWidth()
   {
      return cxRemoteDesktop * zoomPercent / 100;
   }

   public int getScaledHeight()
   {
      return cyRemoteDesktop * zoomPercent / 100;
   }


   // Returns (0,0) if the zoomed remote desktop is bigger than your display;
   // otherwise, returns the pixel coordinates at which you'd need to draw
   // the zoomed remote desktop in order to center it on the display.

   protected int getCenteredLeft()
   {
      int cxScaled = getScaledWidth();
      return (cxScaled < Compat.SCREEN_WIDTH) ? (Compat.SCREEN_WIDTH-cxScaled)/2 : 0;
   }

   protected int getCenteredTop()
   {
      int cyScaled = getScaledHeight();
      return (cyScaled < Compat.SCREEN_HEIGHT) ? (Compat.SCREEN_HEIGHT-cyScaled)/2 : 0;
   }


   // If you've changed zoom level or fullscreen setting, or if the
   // remote desktop's size has just been discovered, then the
   // RemoteView's current position may be illegal.  Adjust it if
   // necessary, to avoid ever rendering off the edge of the world.

   static final int NEVER_REFRESH = 0;
   static final int REFRESH_IF_MOVED = 1;
   static final int ALWAYS_REFRESH = 2;
   static final int DONT_MOVE_CURSOR = 4;

   protected void railViewToDesktop (int mode)
   {
      if (cxRemoteDesktop == 0 || cyRemoteDesktop == 0) {
         return;
      }

      int ixbScreenCenterOld = ixbScreenCenter;
      int iybScreenCenterOld = iybScreenCenter;

      if (getCenteredLeft() != 0) {
         ixbScreenCenter = cxRemoteDesktop/2;
      } else {
         int cxbHalfScreen = getVisibleBufferWidth()/2;
         if (ixbScreenCenter < cxbHalfScreen) {
            ixbScreenCenter = cxbHalfScreen;
         } else if (ixbScreenCenter > cxRemoteDesktop - cxbHalfScreen) {
            ixbScreenCenter = cxRemoteDesktop - cxbHalfScreen;
         }
      }

      if (getCenteredTop() != 0) {
         iybScreenCenter = cyRemoteDesktop/2;
      } else {
         int cybHalfScreen = getVisibleBufferHeight()/2;
         if (iybScreenCenter < cybHalfScreen) {
            iybScreenCenter = cybHalfScreen;
         } else if (iybScreenCenter > cyRemoteDesktop - cybHalfScreen) {
            iybScreenCenter = cyRemoteDesktop - cybHalfScreen;
         }
      }

      // If we had to move the desktop center, repaint.

      if ((mode & DONT_MOVE_CURSOR) == 0) {
         _session.getCursor().railCursorToView(SessionCursor.LIMIT_TO_SCREEN);
      }

      boolean refresh = false;
      if ((mode & ALWAYS_REFRESH) != 0) {
         refresh = true;
      } else if ((mode & REFRESH_IF_MOVED) != 0) {
         if ((ixbScreenCenterOld != ixbScreenCenter) ||
            (iybScreenCenterOld != iybScreenCenter)) {
            refresh = true;
         }
      }
      if (refresh) {
         RemoteScreen screen = getMyRemoteScreen();
         if (screen != null) {
            screen.touch();
         }
      }
   }


   // Returns the global RemoteScreen iff the user is looking
   // at this view; otherwise, returns null.

   protected RemoteScreen getMyRemoteScreen()
   {
      RemoteScreen screen = VncClient.getRemoteScreen();
      if (screen != null) {
         Session session = screen.getSession();
         if (session == _session) {
            return screen;
         }
      }
      return null;
   }


   // Returns the framebuffer for this view.  Never null.

   protected FrameBuffer getMyFrameBuffer()
   {
      return _session.getFrameBuffer();
   }


   Session _session;
}

