package com.randomsoftware.vncclient;

import danger.app.*;
import danger.ui.*;
import danger.util.*;
import danger.crypto.*;
import danger.audio.*;
import java.util.*;

// The Session class represents an attempt to talk to a remote server.
// It manages traffic back and forth, maintains your pseudo-persistent
// state (the RemoteView) and transient state (the FrameBuffer).

public class Session extends Listener
{
   public Session (HostSettings host)
   {
      _host = new HostSettings(host);
      _handshakeState = HandshakeState.WAITING_FOR_SERVER_VERSION;
      _securityTypeCount = 0;
      _verMajor = _verMinor = 0;
      _cursor = new SessionCursor(this);
      _view = new RemoteView(this);
      _frameBuffer = new FrameBuffer(this);
      _connection = new SessionConnection(this);
      _decoder = new ImageDecoder();
      _desktopNameLength = 0;
      _clipboardNeeded = 0;
   }

   public void destroy()
   {
      _frameBuffer.destroy();
      _connection.destroy();
   }

   public void retry()
   {
      if (!SessionState.isTrying (getState())) {
         setHandshakeState (HandshakeState.WAITING_FOR_SERVER_VERSION);
         _securityTypeCount = 0;
         _verMajor = _verMinor = 0;
         _desktopNameLength = 0;
         _clipboardNeeded = 0;
         _decoder.reset();
      }
      _connection.retry();
   }

   public FrameBuffer getFrameBuffer()
   {
      return _frameBuffer;
   }

   public RemoteView getRemoteView()
   {
      return _view;
   }

   public ImageDecoder getImageDecoder()
   {
      return _decoder;
   }

   public SessionCursor getCursor()
   {
      return _cursor;
   }

   public SessionConnection getConnection()
   {
      return _connection;
   }

   public int getState()
   {
      if (_connection == null)
         return SessionState.DISCONNECTED;

      int state = _connection.getState();

      // If the server hung up, it might be something we did
      if (state == SessionState.CONNECTION_LOST) {
         if (_handshakeState == HandshakeState.WAITING_FOR_SECURITY_TYPES) {
            if (isLegalVersion()) {
               state = SessionState.STOP_BUGGING_ME;
            } else {
               state = SessionState.BAD_VERSION;
            }
         } else if (_handshakeState == HandshakeState.WAITING_FOR_SECURITY_RESULT) {
            state = SessionState.BAD_PASSWORD;
         }
      }

      return state;
   }

   public int getHandshakeState()
   {
      return _handshakeState;
   }

   public void setHandshakeState (int newState)
   {
      _handshakeState = newState;

      RemoteScreen screen;
      if ((screen = getMyRemoteScreen()) != null) {
         screen.createStatePopup();
      }
   }

   public HostSettings getHostSettings()
   {
      return _host;
   }

   public String getName()
   {
      return _host.friendlyName;
   }

   public RemoteView getView()
   {
      return _view;
   }

   public void updateHostSettings (HostSettings host)
   {
      _host = new HostSettings(host);
      _view.updateHostSettings(_host);
   }

   public boolean receiveEvent (Event ee)
   {
      int type = Compat.Event_getType(ee);

      if (SessionState.isSessionState (type)) {
         forwardStateChangeEvent(ee);
      }

      switch (type) {
         case SessionConnectionEvent.RECEIVED:
            while (onReceive())
               ;
            return true;
      }

      return super.receiveEvent(ee);
   }

   public void paint (Pen pp)
   {
      _frameBuffer.paint (pp);
      if (getState() == SessionState.CONNECTED) {
         _cursor.paint (pp);
      }
   }

   protected void forwardStateChangeEvent (Event ee)
   {
      VncClient.getHostListScreen().refresh();

      RemoteScreen screen;
      if ((screen = getMyRemoteScreen()) != null) {
         screen.receiveEvent(ee);
      }

      if (getState() == SessionState.CONNECTED) {
         _cursor.sendCursorStatus(SessionCursor.SEND_ALWAYS);
      }
   }

   protected RemoteScreen getMyRemoteScreen()
   {
      return _view.getMyRemoteScreen();
   }

   protected boolean isDigit (int b)
   {
      return (b >= (int)'0' && b <= (int)'9');
   }

   protected boolean onReceive()
   {
      // Once we get disconnected, throw away further input

      if (!SessionState.isTrying(getState())) {
         return false;
      }

      // Otherwise, try to do useful stuff with what we got

      boolean rc = true;
      DataStream stream = _connection.getStream();
      switch (_handshakeState) {

         // First, we expect the server to tell us its highest supported version

         case HandshakeState.WAITING_FOR_SERVER_VERSION:
            Debug.out (this, "received serverVersion");
            rc = onReceiveServerVersion();
            break;

         // Second, we expect the server to tell us all about its security

         case HandshakeState.WAITING_FOR_SECURITY_TYPES:
            Debug.out (this, "received securityType");
            if (_verMajor == 3 && _verMinor == 3) {
               rc = onReceiveServerSecurityTypes3();
            } else {
               rc = onReceiveServerSecurityTypes7();
            }
            break;

         case HandshakeState.WAITING_FOR_SECURITY_DATA:
            Debug.out (this, "received securityData");
            rc = onReceiveServerSecurityData();
            break;

         case HandshakeState.WAITING_FOR_SECURITY_RESULT:
            Debug.out (this, "received securityResult");
            rc = onReceiveServerSecurityResult();
            break;

         // Finally, we need to jump-start the client/server init steps

         case HandshakeState.STARTING_INITIALIZATION:
            Debug.out (this, "received startInit");
            rc = onReceiveStartingInitialization();
            break;

         case HandshakeState.WAITING_FOR_INITIALIZATION:
            Debug.out (this, "received serverInit");
            rc = onReceiveServerInitialization();
            break;

         // Whee!  We're done!

         case HandshakeState.ONLINE:
            while (rc) {
               if (_clipboardNeeded != 0) {
                  if (stream.avail() < _clipboardNeeded) {
                     rc = false;
                     break;
                  }

                  byte[] data = new byte[ _clipboardNeeded ];
                  stream.readBytes (data, 0, _clipboardNeeded);
                  _clipboardNeeded = 0;
                  Pasteboard.setString (new String(data));
               }

               if (_decoder.isActive() && (stream.avail() != 0)) {
                  Debug.out (this, "received continued image data");
                  _decoder.decode(this, _frameBuffer, _connection.getStream());
               }
               if (_decoder.isActive() || stream.avail() == 0) {
                  rc = false;
                  break;
               }

               int msgType = stream.peekInt1();
               switch (msgType) {
                  case Protocol.MESSAGEID_IN_BUFFERUPDATE:
                     Debug.out (this, "received frame buffer update");
                     rc = _decoder.start(this, _connection.getStream());
                     break;

                  case Protocol.MESSAGEID_IN_BELL:
                     Debug.out (this, "received bell request");
                     stream.readInt1();
                     Meta.play (Meta.BEEP_ACTION_FAIL);
                     break;

                  case Protocol.MESSAGEID_IN_SERVER_CLIP:
                     if (stream.avail() < 8) {
                        rc = false;
                     } else {
                        stream.skip (4);
                        _clipboardNeeded = stream.readInt4();
                        Debug.out (this, "received clipboard text ("+_clipboardNeeded+" chars)");
                     }
                     break;

                  default:
                     _connection.getStream().skipAll();
                     Debug.out(this, "received unknown request "+msgType);
                     break;
               }
            }
            break;
      }

      if (_handshakeState != HandshakeState.ONLINE) {
         _connection.processed (SessionConnection.NEED_MORE_DATA);
      } else if (_decoder.lastYieldWasVoluntary()) {
         _connection.processed (SessionConnection.STILL_HAVE_DATA);
      } else {
         _connection.processed (SessionConnection.NEED_MORE_DATA);
      }

      return rc;
   }

   /*** ServerVersion
    *
    * The server sends "RFB 003.003\n", and we send the same thing back (unless,
    * of course, we dislike the version it suggests or vice versa.)
    *
    */

   protected boolean isLegalVersion()
   {
      return (_verMajor == 3) && (_verMinor == 3 || _verMinor == 7 || _verMinor == 8);
   }

   protected boolean onReceiveServerVersion()
   {
      // Did the server send an RFB version-type packet?

      DataStream stream = _connection.getStream();
      if (stream.avail() < 12) {
         return false;
      }

      int a1,a2,a3;
      int b1,b2,b3;
      if (stream.readByte() != 'R' ||
          stream.readByte() != 'F' ||
          stream.readByte() != 'B' ||
          stream.readByte() != ' ' ||
          !isDigit(a1 = stream.readInt1()) ||
          !isDigit(a2 = stream.readInt1()) ||
          !isDigit(a3 = stream.readInt1()) ||
          stream.readByte() != '.' ||
          !isDigit(b1 = stream.readInt1()) ||
          !isDigit(b2 = stream.readInt1()) ||
          !isDigit(b3 = stream.readInt1()) ||
          stream.readByte() != '\n') {
         Debug.out (this, "initial server handshake bad");
         _connection.hangup(SessionState.CONNECT_FAILED);
         return false;
      }

      // So, what version was the server suggesting?

      _verMajor = (int)(a1-'0') * 100 + (int)(a2-'0') * 10 + (int)(a3-'0');
      _verMinor = (int)(b1-'0') * 100 + (int)(b2-'0') * 10 + (int)(b3-'0');
      Debug.out (this, "server reports version "+_verMajor+"."+_verMinor);

      // If the server doesn't at least talk 3.3, walk away

      if (_verMajor < 3 || (_verMajor == 3 && _verMinor < 3)) {
         _connection.hangup(SessionState.BAD_VERSION);
         return false;
      }

      // If the server wants to talk 4.0, ask for 3.8 instead

      if (_verMajor > 3 || (_verMajor == 3 && _verMinor > 8)) {
         _verMajor = 3;
         _verMinor = 8;
      }

      // If the server wants to talk anything between 3.4 and 3.7, use 3.3

      if (_verMajor == 3 && _verMinor < 7) {
         _verMinor = 3;
      }

      // Now tell the server what version we think should be used

      Debug.out (this, "asking server to use version "+_verMajor+"."+_verMinor);
      setHandshakeState(HandshakeState.WAITING_FOR_SECURITY_TYPES);
      Protocol.sendVersionRequest (_connection, _verMajor, _verMinor);
      return true;
   }


   /*** ServerSecurityTypes
    *
    * In v3, the server sends a four-byte int:
    *    0 - server didn't like the version we tried to negotiate
    *    1 - server has no password
    *    2 - server uses DES encryption, so we need to go to SecurityData type
    *
    * In v7, the server sends one byte followed by an array of security options.
    * We'll wait for that whole array and decide if we can proceed.
    *
    */

   static final int SECURITYTYPE_NONE = 1;
   static final int SECURITYTYPE_DES = 2;

   protected boolean onReceiveServerSecurityTypes3()
   {
      DataStream stream = _connection.getStream();
      if (stream.avail() < 4) {
         return false;
      }

      int secType = stream.readInt4();
      if (secType == 0) {
         Debug.out (this, "server refused our version request");
         _connection.hangup(SessionState.BAD_VERSION);
         return false;
      } else if (secType == SECURITYTYPE_NONE) {
         Debug.out (this, "server says unsecured connection is okay");
         setHandshakeState(HandshakeState.STARTING_INITIALIZATION);
         return true;
      } else if (secType == SECURITYTYPE_DES) {
         Debug.out (this, "server says it wants a password");
         setHandshakeState(HandshakeState.WAITING_FOR_SECURITY_DATA);
         onReceiveServerSecurityData();
         return true;
      } else {
         Debug.out (this, "server wants weird security type "+secType);
         _connection.hangup(SessionState.BAD_SECURITY);
         return false;
      }
   }

   protected boolean onReceiveServerSecurityTypes7()
   {
      DataStream stream = _connection.getStream();

      if (_securityTypeCount == 0) {
         if (stream.avail() < 1) {
            return false;
         }
         _securityTypeCount = stream.readInt1();
         if (_securityTypeCount == 0) {
            Debug.out (this, "server refused our version request");
            _connection.hangup(SessionState.BAD_VERSION);
            return false;
         }
      }

      // Get that list of security types

      if (stream.avail() < _securityTypeCount) {
         return false;
      }

      // If we find a type we can support, ask for it and return

      int secChosen = -1;

      for (int ii = 0; ii < _securityTypeCount; ++ii) {
         int secType = stream.readInt1();
         if (secChosen == -1) {
            if (secType == SECURITYTYPE_NONE || secType == SECURITYTYPE_DES) {
               secChosen = secType;
            }
         }
      }

      if (secChosen == SECURITYTYPE_NONE) {
         if (_verMajor == 3 && _verMinor == 8) {
            setHandshakeState(HandshakeState.WAITING_FOR_SECURITY_RESULT);
         } else {
            setHandshakeState(HandshakeState.STARTING_INITIALIZATION);
         }
      } else if (secChosen == SECURITYTYPE_DES) {
         setHandshakeState(HandshakeState.WAITING_FOR_SECURITY_DATA);
      }

      if (secChosen == -1) {
         Debug.out (this, "server wants weird security type");
         _connection.hangup(SessionState.BAD_SECURITY);
         return false;
      } else {
         Protocol.sendInt1 (_connection, secChosen);
         return true;
      }
   }


   /*** ServerSecurityData
    *
    * We only get here if there's a password on the server.  We're
    * waiting for a 64-byte DES key.
    *
    */

   protected boolean onReceiveServerSecurityData()
   {
      DataStream stream = _connection.getStream();
      if (stream.avail() < 16) {
         return false;
      }

      // The DES encoder we have works in 8-byte pieces.
      // So we'll have to feed it two 8-byte-aligned buffers.

      byte[] key1 = new byte[8];
      byte[] key2 = new byte[8];
      stream.readBytes (key1, 0, 8);
      stream.readBytes (key2, 0, 8);

      // What password should we use?

      String password = _host.tempPassword;
      _host.tempPassword = null;
      if (password == null) {
         password = _host.password;
      }
      if (password.length() == 0) {
         _connection.hangup(SessionState.NO_PASSWORD);
         return false;
      }

      // That password is the key; initialize the cipher
      // Then encrypt the challenge, and send it back

      DES des = new DES (DES.hashPassword (password));
      byte[] d1 = des.encrypt(key1);
      byte[] d2 = des.encrypt(key2);

      setHandshakeState(HandshakeState.WAITING_FOR_SECURITY_RESULT);
      Debug.out (this, "sending password \""+password+"\"");

      Protocol.sendPassword (_connection, d1, d2);
      return true;
   }


   /*** ServerSecurityResult
    *
    * Tells us if our password was accepted.  In v3.8, we'll also
    * get this even if there's no password on the server.
    *
    */

   protected boolean onReceiveServerSecurityResult()
   {
      DataStream stream = _connection.getStream();
      if (stream.avail() < 4) {
         return false;
      }
      if (stream.readInt4() != 0) {
         Debug.out (this, "server refused our password");
         _connection.hangup(SessionState.BAD_PASSWORD);
         return false;
      }
      setHandshakeState(HandshakeState.STARTING_INITIALIZATION);
      return true;
   }


   /*** StartingInitialization
    *
    * Tells the server what we want.
    *
    */

   protected boolean onReceiveStartingInitialization()
   {
      // Send the ClientInit message
      Protocol.sendClientInit (_connection, _host.makeExclusive);

      // Send the SetPixelFormat message
      Protocol.sendSetPixelFormat (_connection);

      // Send the SetEncodings message
      Protocol.sendSetEncodings (_connection);

      // Now wait for the server to ack
      setHandshakeState(HandshakeState.WAITING_FOR_INITIALIZATION);
      return false;
   }


   /*** ServerInitialization
    *
    * ...where the server tells us its capabilities
    *
    */

   protected boolean onReceiveServerInitialization()
   {
      DataStream stream = _connection.getStream();
      if (_desktopNameLength == 0) {

         // Phase 1: read the fixed-length part of the initialization packet
         if (stream.avail() < 24) {
            return false;
         }

         // Now we know what the remote desktop is like!  Finally...
         int width = stream.readInt2();
         int height = stream.readInt2();
         stream.skip(16); // Throw away the server's preferred pixel format
         _desktopNameLength = stream.readInt4();
         _view.setRemoteSize (width, height);
         return false;

      } else {

         // Phase 2: read the variable-length part and throw it away
         if (stream.avail() < _desktopNameLength) {
            return false;
         }
         stream.skip(_desktopNameLength);

         setHandshakeState(HandshakeState.ONLINE);
         _connection.setState(SessionState.CONNECTED);

         Protocol.sendClipboard(_connection);
         return false;

      }
   }

   RemoteView _view;
   HostSettings _host;
   SessionCursor _cursor;
   FrameBuffer _frameBuffer;
   SessionConnection _connection;
   ImageDecoder _decoder;
   int _verMajor;
   int _verMinor;
   int _handshakeState;
   int _securityTypeCount;
   int _desktopNameLength;
   int _clipboardNeeded;
}

