package com.randomsoftware.vncclient;

import danger.app.*;
import danger.ui.*;
import danger.util.*;
import java.util.*;
import java.net.*;
import java.io.*;

class SessionConnectionEvent extends SessionState
{
   static final int RECEIVED = 11;
}

// The SessionConnection class maintains an open socket through which
// it communicates with the server.  It also keeps a buffer of received
// data, and notifies the listener that you supply whenever something
// changes.

public class SessionConnection implements Runnable
{
   // UI-Thread Interface

   public SessionConnection (Session ss)
   {
      _everConnected = false;
      _session = ss;
      _state = SessionState.CONNECTING;
      _socket = null;
      _port = 0;
      _incoming = new DataStream();
      _sendThread = new SendThread(this);
      _thread = new Thread (this, "VNC Client Connection Thread");
      _thread.start();
   }

   public synchronized int getState()
   {
      return _state;
   }

   public synchronized void destroy()
   {
      _sendThread.destroy();

      // This will tell the background thread to shut down
      _state = SessionState.DISCONNECTED;
      // ..and this will kill the receive loop
      if (_socket != null) {
         try { _socket.close(); } catch(Exception e) { }
      }
      // ...and this will kill anyone stalled waiting for a
      // repeat performance
      notify();
   }

   public synchronized void hangup (int newState)
   {
      _state = newState;
      _session.sendLowPriorityEvent (new Event (_state));

      if (_socket != null) {
         try { _socket.close(); } catch(Exception e) { }
      }
   }

   public synchronized void setState (int newState)
   {
      _state = newState;
      _session.sendLowPriorityEvent (new Event (_state));
   }

   public synchronized void retry()
   {
      if ((_state != SessionState.DISCONNECTED) &&
          (_state != SessionState.NEGOTIATING) &&
          (_state != SessionState.CONNECTING)) {
         notify();
      }
   }

   static final int NEED_MORE_DATA = 1;
   static final int STILL_HAVE_DATA = 2;
   public synchronized void processed (int urgency)
   {
      if (urgency == NEED_MORE_DATA) {
         notify();
      } else if (_incoming.avail() < RECEIVE_THROTTLE_MIN) {
         notify();
      }
   }

   public DataStream getStream()
   {
      return _incoming;
   }

   public SendThread getSendThread()
   {
      return _sendThread;
   }

   public Socket getSocket()
   {
      return (SessionState.isTrying(_state)) ? _socket : null;
   }

   // Background-Thread Interface

   public void run()
   {
      for(;;) {

         // Get ready to connect.  We'll only quit when someone has
         // called the destroy method, which sets our state to
         // DISCONNECTED (even if the socket is still open).

         synchronized(this) {

            // Don't need this old socket any more
            if (_socket != null) {
               try { _socket.close(); } catch(Exception e) { }
               _socket = null;
            }

            // Throw away any old cached incoming data
            _incoming.skipAll();

            // If we failed to get started last time, wait to get told to retry
            if ((_state != SessionState.CONNECTING) &&
                (_state != SessionState.NEGOTIATING) &&
                (_state != SessionState.DISCONNECTED) &&
                (!_everConnected)) {
               try { wait(); } catch(InterruptedException e) { }
            }

            if (_state == SessionState.DISCONNECTED) {
               break;
            }

            if (_state != SessionState.CONNECTING &&
                _state != SessionState.NEGOTIATING) {
               _state = SessionState.CONNECTING;
            }

            _session.sendLowPriorityEvent (new Event (_state));
         }

         // Split the address into hostname and port.  If the port number
         // is less than 100, it's probably a display number; add 5900
         // to it.  And if no port is specified, just use 5900.

         String host = new String(_session.getHostSettings().serverAddress);
         int port = PORT_DEFAULT;

         int ich;
         if ((ich = host.indexOf(':')) != -1) {
            port = Integer.parseInt (host.substring(ich+1));
            host = host.substring(0,ich);
         }
         if (port < DISPLAY_NUMBER_MAX) {
            port += PORT_DEFAULT;
         }

         // Try to connect

         try {
            _socket = new Socket (host, port);
            if (_socket != null && _state == SessionState.CONNECTING) {
               _state = SessionState.NEGOTIATING;
               _session.sendLowPriorityEvent (new Event (_state));
            }
         } catch(Exception e) {
            _socket = null;
            _state = SessionState.CONNECT_FAILED;
            _session.sendLowPriorityEvent (new Event (_state));
         }

         _sendThread.purge();

         // If we got connected, then we'll sit in a receive loop
         // until someone shuts us down.

         while (SessionState.isTrying(_state)) {
            DataBuffer buf = _incoming.allocateBuffer();

            try {
               InputStream io = _socket.getInputStream();
               int got;
               do {
                  got = io.read(buf.data);
               } while (got == 0);
               if (got > 0) {
                  buf.cbTotal = got;
                  buf.ibFirst = 0;
               } else {
                  _incoming.releaseBuffer(buf);
                  break;
               }
            } catch (Exception e) {
               _incoming.releaseBuffer(buf);
               break;
            }

            synchronized(this) {
               _incoming.append(buf);
               _session.sendLowPriorityEvent (new Event (SessionConnectionEvent.RECEIVED));

               if (_incoming.avail() > RECEIVE_THROTTLE_MAX) {
                  try { wait(); } catch (InterruptedException ee) { }
               }
            }
         }

         synchronized(this) {
            if (SessionState.isTrying(_state)) {
               _state = SessionState.CONNECTION_LOST;
               _session.sendLowPriorityEvent (new Event (_state));
            }
         }
      }
   }

   int _state;
   boolean _everConnected;
   Session _session;
   Socket _socket;
   String _host;
   int _port;
   Thread _thread;
   DataStream _incoming;
   SendThread _sendThread;

   static final int PORT_DEFAULT = 5900;
   static final int DISPLAY_NUMBER_MAX = 100;
   static final int RECEIVE_THROTTLE_MAX = 65535;  // stop receiving until we process some
   static final int RECEIVE_THROTTLE_MIN = 16384;  // start receiving again
}

