package com.randomsoftware.common.animation;

import java.util.*;
import danger.app.*;
import danger.ui.*;
import danger.util.*;

public class AnimWindow extends ScreenWindow
{
   static public final boolean showUpdateAreas = false;
   static public final int framesPerSecond = 20;

   static public final int SMALL_WINDOW = 0;
   static public final int LARGE_WINDOW = 1;

   public AnimWindow (InputListener parent, int size)
   {
      setFullScreen (size == LARGE_WINDOW);

      _parent = parent;
      _rBounds = getBounds();
      _timer = null;
      _timerRefCount = 0;
      _framenumber = 0;
      _sequencer = new AnimSequencer();
      _fullCoverage = false;
      _anyInvalid = true;
   }

   public void show()
   {
      super.show();
      if (_timer != null)
         _timer.start();
   }

   public void hide()
   {
      if (_timer != null)
         _timer.stop();
      super.hide();
   }

   public void startTimer()
   {
      if (_timer == null) {
         _timer = new danger.app.Timer (1000 / framesPerSecond, true, this);
         if (isVisible())
            _timer.start();
      }
   }

   public void stopTimer()
   {
      if (_timer != null) {
         _timer.stop();
         _timer = null;
      }
   }

   public void requestTimer()
   {
      if ((++_timerRefCount) == 1)
         startTimer();
   }

   public void releaseTimer()
   {
      if ((--_timerRefCount) == 0)
         stopTimer();
   }

   public void queueEvent (AnimSequencerEvent event)
   {
      if (_sequencer.isEmpty())
         requestTimer();
      _sequencer.queueEvent (event);
   }

   public void setFullCoverage (boolean f)
   {
      _fullCoverage = f;
   }

   public void touch()
   {
      touch (_rBounds);
   }

   public void touch (Rect rr)
   {
      if (_fullCoverage)
         _anyInvalid = true;
      else
         invalidate (rr);
   }

   public int getNextFrameNumber()
   {
      return _framenumber+1;
   }

   public void finishAnimation()
   {
      if (!_sequencer.isEmpty())
      {
         while (_framenumber < _sequencer.getFinalFrame())
            _sequencer.advanceFrame (++_framenumber);
         if (_sequencer.isEmpty())
            releaseTimer();
      }
   }

   public void purgeSprites()
   {
      _sprites.removeAll();
   }

   public void updateScene()
   {
      // Let the sequencer know we're running

      if (!_sequencer.isEmpty())
      {
         _sequencer.advanceFrame (++_framenumber);
         if (_sequencer.isEmpty())
            releaseTimer();
      }
      else
      {
         ++_framenumber;
      }

      // Tell every Animsprite we're doing a new frame

      if (_fullCoverage)
         invalidate (_rBounds);

      update();

      for (AnimSprite obj = _sprites.getFurthest(); obj != null; obj = _sprites.getCloser(obj))
      {
         obj.update();
      }
   }

   static Random _randgen = new Random();

   public void paintBackground (Pen pp)
   {
      // a good method for you to override
   }

   public void update()
   {
      // a good method for you to override
   }

   public void paint (Pen pp)
   {
      // The damned OS tends to send status-bar updates even if we're full-screen
      if (pp.getClipGlobal().bottom < 160)
      {
         invalidate();  // make sure we get a full repaint instead RSN
         return;
      }

      if (_timer == null)
         updateScene();

      if (!_fullCoverage || _anyInvalid)
      {
         paintBackground (pp);

         for (AnimSprite obj = _sprites.getFurthest(); obj != null; obj = _sprites.getCloser(obj))
         {
            if (obj.isOnScreen())
            {
               pp.push();
               pp.offset (obj.pos.left, obj.pos.top);
               pp.setClip (0, 0, obj.pos.getWidth(), obj.pos.getHeight());
               obj.paint (pp);
               pp.pop();
            }
         }

         _anyInvalid = false;
      }
   }

   public boolean receiveEvent (Event e)
   {
      switch (AnimCompat.Event_getType(e))
      {
         case EventType.EVENT_TIMER:
            updateScene();
            break;
      }
      return (super.receiveEvent(e));
   }

   public boolean eventKeyDown (char c, Event e)
   {
      return (_parent.eventKeyDown(c,e) || super.eventKeyDown(c,e));
   }

   public boolean eventKeyUp (char c, Event e)
   {
      return (_parent.eventKeyUp(c,e) || super.eventKeyUp(c,e));
   }

   public boolean eventKeyRepeat (char c, Event e)
   {
      return (_parent.eventKeyRepeat(c,e) || super.eventKeyRepeat(c,e));
   }

   public boolean eventRawKeyDown (int c, Event e)
   {
      return (_parent.eventRawKeyDown(c,e) || super.eventRawKeyDown(c,e));
   }

   public boolean eventRawKeyUp (int c, Event e)
   {
      return (_parent.eventRawKeyUp(c,e) || super.eventRawKeyUp(c,e));
   }

   public boolean eventWidgetDown (int c, Event e)
   {
      return (_parent.eventWidgetDown(c,e) || super.eventWidgetDown(c,e));
   }

   public boolean eventWidgetUp (int c, Event e)
   {
      return (_parent.eventWidgetUp(c,e) || super.eventWidgetUp(c,e));
   }

   public boolean eventWidgetHeld (int c, Event e)
   {
      return (_parent.eventWidgetHeld(c,e) || super.eventWidgetHeld(c,e));
   }

   void removeSprite (AnimSprite o)
   {
      if (!_sequencer.isEmpty())
      {
         _sequencer.removeSprite(o);
         if (_sequencer.isEmpty())
            releaseTimer();
      }
   }

   protected Rect _rBounds;
   protected InputListener _parent;
   protected AnimSpriteList _sprites = new AnimSpriteList();
   protected danger.app.Timer _timer;
   protected int _timerRefCount;
   protected AnimSequencer _sequencer;
   protected int _framenumber;
   protected boolean _fullCoverage;
   protected boolean _anyInvalid;
}

