package com.randomsoftware.common.meshlib;
import com.randomsoftware.common.animation.*;

import java.util.*;
import danger.app.*;
import danger.ui.*;
import danger.util.*;
import danger.audio.*;

public class Camera extends AnimSprite
{
   public static long DEFAULT_NEAR_PLANE = 0-FixedMath.PREC*2;
   public static long DEFAULT_FAR_PLANE = 1000 << FixedMath.SHIFT;
   public static long DEFAULT_FOG_NEAR_PLANE = 10 << FixedMath.SHIFT;
   public static long DEFAULT_FOG_FAR_PLANE = 200 << FixedMath.SHIFT;

   // We don't support texture mapping or gouraud shading here; sorry.
   // The hardware platform just doesn't have the CPU for doing per-pixel
   // calculations.  Lighting and backface culling can both be turned off,
   // as they as they require calculation of face normals and therefore
   // are computationally expensive.

   public static final int QUALITY_POINT = 1;
   public static final int QUALITY_LINE = 2;
   public static final int QUALITY_SOLID = 3;
   public static final int QUALITY_SOLID_OUTLINED = 4;
   public static final int QUALITY_TEXTURED = 5;
   public static final int QUALITY_TEXTURED_OUTLINED = 6;
   public int quality = QUALITY_TEXTURED;

   public static final int LIGHTING_NONE = 0;
   public static final int LIGHTING_FLAT = 1;
   public int lighting = LIGHTING_NONE;

   public static final int CULL_NONE = 0;
   public static final int CULL_BACKFACE = 1;
   public int cull = CULL_NONE;

   public boolean cullFov = true;
   public long cullFovAngle = FixedMath.HEXIAN_90 * 7 / 9;
   public long cullFovDist = 3 << FixedMath.SHIFT;

   public static final int FOG_NONE = 0;
   public static final int FOG_LINEAR = 1;
   public int fog = FOG_NONE;
   public Color fogColor = new Color(0,0,0);
   public long fogNearPlane = DEFAULT_FOG_NEAR_PLANE;
   public long fogFarPlane = DEFAULT_FOG_FAR_PLANE;

   public Camera nextCameraInScene;
   public Camera prevCameraInScene;

   public Camera (Scene sc, AnimWindow win, int depth)
   {
      super (win, depth);

      _scene = sc;
      _vSource = new Triplet (0, 0, 0);
      _vTarget = new Triplet (0, 0, FixedMath.PREC);
      _vCam = new Triplet();
      _yAxis = new Triplet (0, FixedMath.PREC, 0);
      _vRight = new Triplet();
      _vUp = new Triplet();
      _mxProj = new FixedMatrix();
      _mxView = new FixedMatrix();
      _mxMultiple = new FixedMatrix();

      _angFOV = 0;
      _dNear = _dFar = 0;
      setProjection (FixedMath.HEXIAN_90, DEFAULT_NEAR_PLANE, DEFAULT_FAR_PLANE);

      // Insert(scene.cameras)
      if ((prevCameraInScene = _scene.lastCamera) != null)
         prevCameraInScene.nextCameraInScene = this;
      else
         _scene.firstCamera = this;
      nextCameraInScene = null;
      _scene.lastCamera = this;
   }

   public void destroy()
   {
      // Remove(scene.cameras)
      if (nextCameraInScene != null)
         nextCameraInScene.prevCameraInScene = prevCameraInScene;
      else
         _scene.lastCamera = prevCameraInScene;
      if (prevCameraInScene != null)
         prevCameraInScene.nextCameraInScene = nextCameraInScene;
      else
         _scene.firstCamera = nextCameraInScene;

      super.destroy();
   }

   public Triplet getSource()
   {
      return _vSource;
   }

   public void setSource (Triplet src)
   {
      if (!_vSource.equals(src))
      {
         _vSource.dx = src.dx;
         _vSource.dy = src.dy;
         _vSource.dz = src.dz;
         _vSource.dw = src.dw;
         buildViewMatrix();
      }
   }

   public Triplet getTarget()
   {
      return _vTarget;
   }

   public void setTarget (Triplet tgt)
   {
      if (!_vTarget.equals(tgt))
      {
         _vTarget.dx = tgt.dx;
         _vTarget.dy = tgt.dy;
         _vTarget.dz = tgt.dz;
         _vTarget.dw = tgt.dw;
         buildViewMatrix();
      }
   }

   public void setOrthgraphic()
   {
      _angFOV = 0;
      _dNear = 0;
      _dFar = 0;

      _mxProj.makeIdentity();

      buildViewMatrix();
   }

   public void setProjection (long angFOV, long dNear, long dFar)
   {
      // Can't approach 180' without getting an infinite width viewing
      // plane, which we'd discover as a divide-by-zero in tangent().

      _angFOV = angFOV;
      if (angFOV > FixedMath.HEXIAN_180 - 10)
         angFOV = FixedMath.HEXIAN_180 - 10;

      _dNear = dNear;
      _dFar = dFar;
      if (_dFar < _dNear+FixedMath.PREC)
         _dFar = _dNear+FixedMath.PREC;

      // Find the width and height of the near clipping plane.  Those
      // are figured based on the sprite's size on screen, and the FOV.

      long cx = pos.right - pos.left;
      long cy = pos.bottom - pos.top;
      _cxyProj = ((cx < cy) ? cx : cy) << (FixedMath.SHIFT-1);

      if (_cxyProj <= 0)
      {
         // The sprite will resize before we draw, and we'll come in here
         // again when that happens. If we don't resize first, then it's
         // irrelevant what we draw because the viewport is 0x0.
         _mxProj.makeIdentity();
      }
      else
      {
         long f = FixedMath.divide (_cxyProj, FixedMath.tangent(_angFOV/2));

         _mxProj.val00 = f; _mxProj.val01 = 0; _mxProj.val02 = 0; _mxProj.val03 = 0;
         _mxProj.val10 = 0; _mxProj.val11 = f; _mxProj.val12 = 0; _mxProj.val13 = 0; 
         _mxProj.val20 = 0; _mxProj.val21 = 0; _mxProj.val22 = FixedMath.PREC; _mxProj.val23 = 0;
         _mxProj.val30 = 0; _mxProj.val31 = 0; _mxProj.val32= -FixedMath.PREC; _mxProj.val33 = FixedMath.PREC;
      }

      buildViewMatrix();
   }

   public boolean isProjection()
   {
      return (_dNear != _dFar);
   }

   public long getFieldOfView()
   {
      return _angFOV;
   }

   public void setRect (Rect rr)
   {
      super.setRect (rr);
      if (isProjection())
         setProjection (_angFOV, _dNear, _dFar);
      else
         buildViewMatrix();
   }

   protected void buildViewMatrix()
   {
      // We want a *reversed* transformation here: we're not actually moving
      // a camera to the right, instead we're moving everything else to the
      // left.  The same applies for rotations: instead of rotating the camera
      // along X+, we're actually rotating everything else along X-.  So
      // what we're encoding is really a transposed camera rotation multiplied
      // by a negative camera translation.  Remember that we use right-handed
      // coordinates here, and expect faces wound counter-clockwise.

      _vCam.dx = _vTarget.dx - _vSource.dx;
      _vCam.dy = _vTarget.dy - _vSource.dy;
      _vCam.dz = _vTarget.dz - _vSource.dz;
      _vCam.dw = _vSource.dw;
      _vCam.normalize();

      _vRight.makeNormalizedCross (_vCam, _yAxis);
      _vUp.makeCross (_vRight, _vCam);

      _mxView.val00 = _vRight.dx;
      _mxView.val01 = _vRight.dy;
      _mxView.val02 = _vRight.dz;
      _mxView.val03 = - (((_vRight.dx*_vSource.dx) + (_vRight.dy*_vSource.dy) + (_vRight.dz*_vSource.dz)) >> FixedMath.SHIFT);

      _mxView.val10 = _vUp.dx;
      _mxView.val11 = _vUp.dy;
      _mxView.val12 = _vUp.dz;
      _mxView.val13 = - (((_vUp.dx*_vSource.dx) + (_vUp.dy*_vSource.dy) + (_vUp.dz*_vSource.dz)) >> FixedMath.SHIFT);

      _mxView.val20 = - _vCam.dx;
      _mxView.val21 = - _vCam.dy;
      _mxView.val22 = - _vCam.dz;
      _mxView.val23 = ((_vCam.dx*_vSource.dx) + (_vCam.dy*_vSource.dy) + (_vCam.dz*_vSource.dz)) >> FixedMath.SHIFT;

      _mxView.val30 = 0;
      _mxView.val31 = 0;
      _mxView.val32 = 0;
      _mxView.val33 = FixedMath.PREC;

      _mxMultiple.makeMultiply (_mxProj, _mxView);

      _scene.invalidate();
      touch();
   }

   public void update()
   {
      _scene.update();
   }

   public void paint (Pen pp)
   {
      pp.setClipGlobal (TriangleClipper.LEFT, TriangleClipper.TOP, TriangleClipper.RIGHT, TriangleClipper.BOTTOM);
      _scene.paint (pp, this);
   }

   public FixedMatrix getViewMatrix()
   {
      return _mxView;
   }

   public FixedMatrix getProjectionMatrix()
   {
      return _mxProj;
   }

   protected Scene _scene;
   protected Triplet _vSource;
   protected Triplet _vTarget;
   protected Triplet _vCam;
   protected Triplet _yAxis;
   protected Triplet _vRight;
   protected Triplet _vUp;
   protected FixedMatrix _mxProj;
   protected FixedMatrix _mxView;
   protected FixedMatrix _mxMultiple;
   protected long _angFOV;
   protected long _dNear;
   protected long _dFar;
   protected long _cxyProj;
}

