package com.randomsoftware.common.meshlib;

import java.util.*;
import danger.app.*;
import danger.ui.*;
import danger.util.*;
import danger.audio.*;

public class FaceSort
{
   FaceSortable[] sortedFaces;
   int sortedFacesInUse;
   int sortedFacesCapacity;

   public FaceSort()
   {
      sortedFaces = null;
      sortedFacesInUse = 0;
      sortedFacesCapacity = 0;
   }

   public void update (Scene sc, boolean fRenderAll)
   {
      sortedFacesInUse = 0;

      // Get a rough estimate of the number of entries we'll need to store.
      // Remember: we're sorting *all* faces in the scene, not just the faces
      // within a particular mesh.

      int cNeeded = sc.countTriggers;
      for (Mesh mm = sc.firstMesh; mm != null; mm = mm.nextMeshInScene)
      {
         if (!mm._fWillRender)
            continue;
         cNeeded += mm.countFaces;
      }
      if (cNeeded == 0)
         return;

      // Make sure our sort array is big enough

      if (cNeeded > sortedFacesCapacity)
      {
         sortedFaces = new FaceSortable [cNeeded];
         sortedFacesCapacity = cNeeded;
      }

      // Then populate the array with raw data

      long dzPivot = 0;

      for (Mesh mm = sc.firstMesh; mm != null; mm = mm.nextMeshInScene)
      {
         if (!mm._fWillRender)
            continue;
         for (Face face = mm.firstFace; face != null; face = face.nextFaceInMesh)
         {
            if (!fRenderAll && !face._fWillRender)
               continue;

            face._dCamera = face.vertex1._dCamera +
                            face.vertex2._dCamera +
                            face.vertex3._dCamera;

            dzPivot += face._dCamera;
            sortedFaces[ sortedFacesInUse++ ] = face;
         }
      }

      for (Trigger trigger = sc.firstTrigger; trigger != null; trigger = trigger.nextTriggerInScene)
      {
         dzPivot += trigger._dCamera;
         sortedFaces[ sortedFacesInUse++ ] = trigger;
      }

      // Now sort that array.

      if (sortedFacesInUse != 0)
      {
         quicksort (0, sortedFacesInUse-1, dzPivot/sortedFacesInUse);
      }
   }

   boolean belongsLeft (int idx, long dz)
   {
      return (sortedFaces[idx]._dCamera >= dz);
   }

   boolean isInOrder (int idxA, int idxB)
   {
      return (sortedFaces[idxA]._dCamera >= sortedFaces[idxB]._dCamera);
   }

   void swap (int idxA, int idxB)
   {
      FaceSortable ff = sortedFaces[idxA];
      sortedFaces[idxA] = sortedFaces[idxB];
      sortedFaces[idxB] = ff;
   }

   void quicksort (int iFirst, int iLast, long dzPivot)
   {
      if (iFirst+4 > iLast)
      {
         if (iFirst+1 == iLast)       // A,B
         {
            if (!isInOrder (iFirst, iLast))
               swap (iFirst, iLast);
         }
         else if (iFirst+2 == iLast)  // A,B,C
         {
            if (!isInOrder (iFirst, iFirst+1))
               swap (iFirst, iFirst+1);
            if (!isInOrder (iFirst, iFirst+2))
               swap (iFirst, iFirst+2);
            if (!isInOrder (iFirst+1, iFirst+2))
               swap (iFirst+1, iFirst+2);
         }
         else if (iFirst+3 == iLast)  // A,B,C,D
         {
            if (!isInOrder (iFirst, iFirst+1))
               swap (iFirst, iFirst+1);
            if (!isInOrder (iFirst+2, iFirst+3))
               swap (iFirst+2, iFirst+3);
            if (!isInOrder (iFirst+1, iFirst+2))
               swap (iFirst+1, iFirst+2);
            if (!isInOrder (iFirst, iFirst+1))
               swap (iFirst, iFirst+1);
            if (!isInOrder (iFirst+2, iFirst+3))
               swap (iFirst+2, iFirst+3);
         }
      }
      else if (iFirst < iLast)
      {
         long dzPivotLeft = 0;
         long dzPivotRight = 0;

         int iWriteLeft = iFirst;
         int iWriteRight = iLast;
         while (iWriteLeft <= iWriteRight)
         {
            if (belongsLeft (iWriteLeft, dzPivot))
            {
               dzPivotLeft += sortedFaces[iWriteLeft]._dCamera;
               ++iWriteLeft;
            }
            else
            {
               while ((iWriteRight >= iWriteLeft) && (!belongsLeft (iWriteRight, dzPivot)))
               {
                  dzPivotRight += sortedFaces[iWriteRight]._dCamera;
                  --iWriteRight;
               }

               if ((iWriteRight >= iWriteLeft) && (belongsLeft (iWriteRight, dzPivot)))
               {
                  swap (iWriteLeft, iWriteRight);
                  dzPivotLeft += sortedFaces[iWriteLeft]._dCamera;
                  ++iWriteLeft;
               }
            }
         }

         if ((iWriteLeft != iFirst) && (iWriteRight != iLast))
         {
            if (iWriteLeft > iFirst)
               quicksort (iFirst, iWriteLeft-1, dzPivotLeft/(iWriteLeft-iFirst));
            if (iWriteRight < iLast)
               quicksort (iWriteRight+1, iLast, dzPivotRight/(iLast-iWriteRight));
         }
      }
   }
}

