package com.randomsoftware.common.meshlib;

import java.util.*;
import danger.app.*;
import danger.ui.*;
import danger.util.*;
import danger.audio.*;

public class Mesh extends Entity
{
   public Vertex firstVertex = null;
   public Vertex lastVertex = null;

   public Texture firstTexture = null;
   public Texture lastTexture = null;

   public Face firstFace = null;
   public Face lastFace = null;
   public int countFaces = 0;

   public Line firstLine = null;
   public Line lastLine = null;

   public boolean isUnlit = false;

   public Mesh()
   {
      _scene = null;
      _actor = null;
   }

   public Mesh (MeshResource mr)
   {
      _scene = null;
      _actor = null;
      load (mr);
   }

   public Mesh (Scene scene)
   {
      connect (scene);
   }

   public Mesh (Actor actor)
   {
      connect (actor);
   }

   public Mesh (Scene scene, MeshResource mr)
   {
      connect (scene);
      load (mr);
   }

   public Mesh (Actor actor, MeshResource mr)
   {
      connect (actor);
      load (mr);
   }

   public void destroy()
   {
      disconnect();
      invalidate();

      while (firstFace != null)
      {
         firstFace.destroy();
         firstFace = firstFace.nextFaceInMesh;
      }
      firstFace = lastFace = null;
   }

   public void connect (Scene scene)
   {
      _scene = scene;
      _actor = null;

      // Insert(scene.meshes)
      if ((prevMeshInScene = _scene.lastMesh) != null)
         prevMeshInScene.nextMeshInScene = this;
      else
         _scene.firstMesh = this;
      nextMeshInScene = null;
      _scene.lastMesh = this;
   }

   public void connect (Actor actor)
   {
      _scene = actor._scene;
      _actor = actor;

      // Insert(scene.meshes)
      if ((prevMeshInScene = _scene.lastMesh) != null)
         prevMeshInScene.nextMeshInScene = this;
      else
         _scene.firstMesh = this;
      nextMeshInScene = null;
      _scene.lastMesh = this;

      // Insert(actor.meshes)
      if ((prevMeshInActor = _actor.lastMesh) != null)
         prevMeshInActor.nextMeshInActor = this;
      else
         _actor.firstMesh = this;
      nextMeshInActor = null;
      _actor.lastMesh = this;
   }

   public void disconnect()
   {
      if (_scene != null)
      {
         // Remove(scene.meshes)
         if (nextMeshInScene != null)
            nextMeshInScene.prevMeshInScene = prevMeshInScene;
         else
            _scene.lastMesh = prevMeshInScene;
         if (prevMeshInScene != null)
            prevMeshInScene.nextMeshInScene = nextMeshInScene;
         else
            _scene.firstMesh = nextMeshInScene;
         _scene = null;
      }

      if (_actor != null)
      {
         // Remove(actor.meshes)
         if (nextMeshInActor != null)
            nextMeshInActor.prevMeshInActor = prevMeshInActor;
         else
            _actor.lastMesh = prevMeshInActor;
         if (prevMeshInActor != null)
            prevMeshInActor.nextMeshInActor = nextMeshInActor;
         else
            _actor.firstMesh = nextMeshInActor;
         _actor = null;
      }
   }

   public void load (MeshResource mr)
   {
      if (mr._faces == null)
         return;

      for (int ii = 0; ii < mr._textures.length; ++ii)
      {
         mr._textures[ii].texture = addTexture (mr._textures[ii]);
      }

      for (int ii = 0; ii < mr._vertices.length; ++ii)
      {
         VertexData vd = mr._vertices[ii];
         mr._vertices[ii].vertex = addVertex (vd.dx, vd.dy, vd.dz);
      }

      for (int ii = 0; ii < mr._faces.length; ++ii)
      {
         addFace (mr._textures, mr._vertices, mr._faces[ii]);
      }
   }

   public Vertex addVertex (long dx, long dy, long dz)
   {
      Vertex vv = new Vertex (dx, dy, dz);
      connectVertex(vv);
      return vv;
   }

   public void removeVertex (Vertex vv)
   {
      disconnectVertex(vv);
   }

   public void connectVertex (Vertex vv)
   {
      // vv.Insert(vertices)
      if ((vv.prevVertexInMesh = lastVertex) != null)
         vv.prevVertexInMesh.nextVertexInMesh = vv;
      else
         firstVertex = vv;
      vv.nextVertexInMesh = null;
      lastVertex = vv;
   }

   public void disconnectVertex (Vertex vv)
   {
      // vv.Remove(vertices)
      if (vv.nextVertexInMesh != null)
         vv.nextVertexInMesh.prevVertexInMesh = vv.prevVertexInMesh;
      else
         lastVertex = vv.prevVertexInMesh;
      if (vv.prevVertexInMesh != null)
         vv.prevVertexInMesh.nextVertexInMesh = vv.nextVertexInMesh;
      else
         firstVertex = vv.nextVertexInMesh;
   }

   public Face addFace (Vertex v1, Vertex v2, Vertex v3, Texture tx)
   {
      Face ff = new Face (_scene, this, v1, v2, v3, tx);
      connectFace(ff);
      return ff;
   }

   public Face addFace (TextureData[] td, VertexData[] vd, FaceData fd)
   {
      Face ff = new Face (_scene, this, td, vd, fd);
      connectFace (ff);
      return ff;
   }

   public void removeFace (Face ff)
   {
      disconnectFace (ff);
      ff.destroy();
   }

   public void connectFace (Face ff)
   {
      // ff.Insert(faces)
      if ((ff.prevFaceInMesh = lastFace) != null)
         ff.prevFaceInMesh.nextFaceInMesh = ff;
      else
         firstFace = ff;
      ff.nextFaceInMesh = null;
      lastFace = ff;

      ff._scene = _scene;
      ff._mesh = this;
      ++countFaces;
   }

   public void disconnectFace (Face ff)
   {
      ff.clearLights();

      // ff.Remove(faces)
      if (ff.nextFaceInMesh != null)
         ff.nextFaceInMesh.prevFaceInMesh = ff.prevFaceInMesh;
      else
         lastFace = ff.prevFaceInMesh;
      if (ff.prevFaceInMesh != null)
         ff.prevFaceInMesh.nextFaceInMesh = ff.nextFaceInMesh;
      else
         firstFace = ff.nextFaceInMesh;

      ff._scene = null;
      ff._mesh = null;
      --countFaces;
   }

   public Line addLine (Vertex v1, Vertex v2, Texture tex)
   {
      Line ll = new Line (v1, v2, tex);

      // ll.Insert(lines)
      if ((ll.prevLineInMesh = lastLine) != null)
         ll.prevLineInMesh.nextLineInMesh = ll;
      else
         firstLine = ll;
      ll.nextLineInMesh = null;
      lastLine = ll;

      return ll;
   }

   public void removeLine (Line ll)
   {
      // ll.Remove(lines)
      if (ll.nextLineInMesh != null)
         ll.nextLineInMesh.prevLineInMesh = ll.prevLineInMesh;
      else
         lastLine = ll.prevLineInMesh;
      if (ll.prevLineInMesh != null)
         ll.prevLineInMesh.nextLineInMesh = ll.nextLineInMesh;
      else
         firstLine = ll.nextLineInMesh;
   }

   public Texture addTexture (TextureData td)
   {
      Texture tex = new Texture (td);

      // tex.Insert(textures)
      if ((tex.prevTextureInMesh = lastTexture) != null)
         tex.prevTextureInMesh.nextTextureInMesh = tex;
      else
         firstTexture = tex;
      tex.nextTextureInMesh = null;
      lastTexture = tex;

      return tex;
   }

   public Texture addTexture (int r, int g, int b, ImageList imageList, String name)
   {
      Texture tex = new Texture (r, g, b, imageList, name);

      // tex.Insert(textures)
      if ((tex.prevTextureInMesh = lastTexture) != null)
         tex.prevTextureInMesh.nextTextureInMesh = tex;
      else
         firstTexture = tex;
      tex.nextTextureInMesh = null;
      lastTexture = tex;

      return tex;
   }

   public void removeTexture (Texture tex)
   {
      // tex.Remove(texturess)
      if (tex.nextTextureInMesh != null)
         tex.nextTextureInMesh.prevTextureInMesh = tex.prevTextureInMesh;
      else
         lastTexture = tex.prevTextureInMesh;
      if (tex.prevTextureInMesh != null)
         tex.prevTextureInMesh.nextTextureInMesh = tex.nextTextureInMesh;
      else
         firstTexture = tex.nextTextureInMesh;
   }

   public void clearLights()
   {
      for (Face face = firstFace; face != null; face = face.nextFaceInMesh)
         face.clearLights();
   }

   public static final int MESH_INVALIDATE_DEFORM   = 0x0001;  // mesh moved or deformed
   public static final int MESH_INVALIDATE_LIGHTS   = 0x0002;  // mesh lighting is affected
   public static final int MESH_INVALIDATE_CAMERA   = 0x0004;  // position in camera adjusted

   public void invalidate()
   {
      if (isUnlit)
         invalidate (MESH_INVALIDATE_CAMERA);
      else
         invalidate (MESH_INVALIDATE_CAMERA | MESH_INVALIDATE_LIGHTS | MESH_INVALIDATE_DEFORM);
   }

   public void invalidate (int flags)
   {
      if (flags == MESH_INVALIDATE_CAMERA)
      {
         super.invalidate();
         _fNeedsCameraPos = true;
      }
      else if ((flags != 0) && (!isChanged))
      {
         super.invalidate();
         _fNeedsRelight = !isUnlit;
         _fNeedsCameraPos = true;
      }
   }

   public static final int MESH_NEED_WORLD_FACENORMALS    = 0x0001;
   public static final int MESH_NEED_CAMERA_FACENORMALS   = 0x0002;
   public static final int MESH_NEED_CAMERA_VERTEXPOS     = 0x0004;

   public void recalculate (Camera cam, FixedMatrix mxWorldToCamera, int dwNeed)
   {
      FixedMatrix mxMeshToWorld = (_actor != null) ? _mxMeshToWorld : this;
      _mxMeshToCamera.makeMultiply (mxWorldToCamera, mxMeshToWorld);

      _dToCamera = 0 - _mxMeshToCamera.val23;
      if (_dToCamera < cam._dNear)
      {
         _fWillRender = false;
      }
      else if ((cam.fog != Camera.FOG_NONE) && (_dToCamera > cam.fogFarPlane))
      {
         _fWillRender = false;
      }
      else if ((!cam.cullFov) || (_dToCamera < cam.cullFovDist))
      {
         _fWillRender = true;
      }
      else
      {
         long xAbs = (_mxMeshToCamera.val03 < 0) ? (0 - _mxMeshToCamera.val03) : (_mxMeshToCamera.val03);
         _fWillRender = (FixedMath.arctangent (FixedMath.divide (xAbs, cam._cxyProj), _dToCamera) < cam.cullFovAngle);
      }

      if (_fWillRender)
      {
         // Find out where each vertex lives in camera space

         if ((dwNeed & (MESH_NEED_CAMERA_FACENORMALS | MESH_NEED_CAMERA_VERTEXPOS)) != 0)
         {
            for (Vertex vv = firstVertex; vv != null; vv = vv.nextVertexInMesh)
            {
               vv._vPosInCamera.makeHomogenizedMultiply (_mxMeshToCamera, vv);
               vv._dCamera = 0 - vv._vPosInCamera.dz;
            }
         }

         // Find the face normals in worldspace

         if ((dwNeed & MESH_NEED_WORLD_FACENORMALS) != 0)
         {
            for (Vertex vv = firstVertex; vv != null; vv = vv.nextVertexInMesh)
            {
               vv._vPosInWorld.makeMultiply (mxMeshToWorld, vv);
            }

            for (Face ff = firstFace; ff != null; ff = ff.nextFaceInMesh)
            {
               t2.dx = ff.vertex2._vPosInWorld.dx - ff.vertex1._vPosInWorld.dx;
               t2.dy = ff.vertex2._vPosInWorld.dy - ff.vertex1._vPosInWorld.dy;
               t2.dz = ff.vertex2._vPosInWorld.dz - ff.vertex1._vPosInWorld.dz;

               t3.dx = ff.vertex3._vPosInWorld.dx - ff.vertex1._vPosInWorld.dx;
               t3.dy = ff.vertex3._vPosInWorld.dy - ff.vertex1._vPosInWorld.dy;
               t3.dz = ff.vertex3._vPosInWorld.dz - ff.vertex1._vPosInWorld.dz;

               if (ff._vNormalInWorld == null)
                  ff._vNormalInWorld = new Triplet();
               ff._vNormalInWorld.makeNormalizedCross (t2, t3);
            }
         }

         // Find the face normals in cameraspace.  There's a real kludge here:
         // our camera transform is interested in dx and dy in order to find a
         // vertex's position on-screen; it doesn't care much about dz, and
         // in fact tends to squash it mercilessly towards zero.  We therefore
         // magnify differences in DZ here by an arbitrary amount, in order to
         // better show dz in the camera-space normals.

         if ((dwNeed & MESH_NEED_CAMERA_FACENORMALS) != 0)
         {
            for (Face ff = firstFace; ff != null; ff = ff.nextFaceInMesh)
            {
               t2.dx = ff.vertex2._vPosInCamera.dx - ff.vertex1._vPosInCamera.dx;
               t2.dy = ff.vertex2._vPosInCamera.dy - ff.vertex1._vPosInCamera.dy;
               t2.dz = (ff.vertex2._vPosInCamera.dz - ff.vertex1._vPosInCamera.dz) * 1000;

               t3.dx = ff.vertex3._vPosInCamera.dx - ff.vertex1._vPosInCamera.dx;
               t3.dy = ff.vertex3._vPosInCamera.dy - ff.vertex1._vPosInCamera.dy;
               t3.dz = (ff.vertex3._vPosInCamera.dz - ff.vertex1._vPosInCamera.dz) * 1000;

               if (ff._vNormalInCamera == null)
                  ff._vNormalInCamera = new Triplet();
               ff._vNormalInCamera.makeNormalizedCross (t2, t3);
            }
         }

         isChanged = false;
         _fNeedsCameraPos = false;
      }
   }

   public void update()
   {
   }

   public Mesh nextMeshInScene;
   public Mesh prevMeshInScene;

   public Mesh nextMeshInActor;
   public Mesh prevMeshInActor;

   static Triplet t2 = new Triplet();
   static Triplet t3 = new Triplet();
   protected Scene _scene = null;
   protected Actor _actor = null;
   protected FixedMatrix _mxMeshToWorld = new FixedMatrix();
   protected FixedMatrix _mxMeshToCamera = new FixedMatrix();
   protected boolean _fNeedsRelight = true;
   protected boolean _fNeedsCameraPos = true;
   protected boolean _fWillRender = true;
   protected long _dToCamera = 0;
}

