package com.randomsoftware.common.meshlib;

import java.util.*;
import java.io.*;
import danger.app.*;
import danger.ui.*;
import danger.util.*;
import danger.audio.*;

public class MeshResource
{
   public static final int RESTYPE_MESH = 1024;

   public MeshResource (ResourceDatabase db, int idr, ImageList imageList)
   {
      _images = null;
      _textures = null;
      _vertices = null;
      _faces = null;

      Resource res = db.getResource(RESTYPE_MESH, idr);
      DataInputStream ds = new DataInputStream (res.getInputStream());

      try {
         read(ds, imageList);
      } catch (IOException e) {
         _textures = null;
         _vertices = null;
         _faces = null;
      }
   }

   void read (DataInputStream ds, ImageList imageList) throws IOException
   {
      if (ds.readInt() != JMC_SIGNATURE_FILE_FIXED_32)
         fail ("mesh must be exported using JMC, using 32-bit fixed-point big-endian binary format.");

      int val;
      boolean haveTextures = false;
      while ((val = ds.readInt()) != JMC_SIG32_END)
      {
         if ((val & (~0xFF)) == JMC_SIG32_IMAGES)
         {
            haveTextures = true;
            readImages(ds, (val & 0xFF));
         }
         else if ((val & (~0xFF)) == JMC_SIG32_TEXTURES)
         {
            readTextures(ds, (val & 0xFF), haveTextures, imageList);
         }
         else if ((val & (~0xFF)) == JMC_SIG32_VERTICES)
         {
            readVertices(ds, (val & 0xFF));
         }
         else if ((val & (~0xFF)) == JMC_SIG32_FACES)
         {
            readFaces(ds, (val & 0xFF), haveTextures);
         }
         else
         {
            fail ("unexpected signature");
         }
      }
   }

   void readImages (DataInputStream ds, int count) throws IOException
   {
      if (_images != null)
         fail ("multiple image sections specified");

      if (count != 0)
      {
         _images = new String[count];
         for (int ii = 0; ii < count; ++ii)
            _images[ii] = popString (ds);
      }
   }

   void readTextures (DataInputStream ds, int count, boolean haveTextures, ImageList imageList) throws IOException
   {
      if (_textures != null)
         fail ("multiple color sections specified");

      _textures = new TextureData[(int)count];
      for (int ii = 0; ii < count; ++ii)
         _textures[ii] = new TextureData(ds, haveTextures, _images, imageList);
   }

   void readVertices (DataInputStream ds, int count) throws IOException
   {
      if (_vertices != null)
         fail ("multiple vertex sections specified");

      _vertices = new VertexData[count];
      for (int ii = 0; ii < count; ++ii)
         _vertices[ii] = new VertexData(ds);
   }

   void readFaces (DataInputStream ds, int count, boolean haveTextures) throws IOException
   {
      if (_faces != null)
         fail ("multiple face sections specified");
      if (_textures == null)
         fail ("face section specified before textures section");
      if (_vertices == null)
         fail ("face section specified before vertex section");

      _faces = new FaceData[count];
      for (int ii = 0; ii < count; ++ii)
      {
         _faces[ii] = new FaceData(ds, haveTextures, _textures);

         if ( (_faces[ii].iVertex1 == _faces[ii].iVertex2) ||
              (_faces[ii].iVertex1 == _faces[ii].iVertex3) ||
              (_faces[ii].iVertex2 == _faces[ii].iVertex3) )
            fail("duplicate vertices detected for face "+ii);

         if ( (_faces[ii].iVertex1 >= _vertices.length) ||
              (_faces[ii].iVertex2 >= _vertices.length) ||
              (_faces[ii].iVertex3 >= _vertices.length) )
            fail("illegal vertex referenced by face "+ii);

         if (_faces[ii].iTexture >= _textures.length)
            fail("illegal texture referenced by face "+ii);
      }
   }

   void fail (String ss) throws IOException
   {
      DEBUG.p (ss);
      throw new IOException();
   }

   public TextureData getTextureData (int ii)
   {
      return _textures[ii];
   }

   public VertexData getVertexData (int ii)
   {
      return _vertices[ii];
   }

   static String popString (DataInputStream ds) throws IOException
   {
      int byteCount = (int)ds.readByte();
      byte[] rawtext = new byte[ byteCount ];
      ds.read (rawtext, 0, byteCount);
      return new String (rawtext);
   }

   String[] _images;
   TextureData[] _textures;
   VertexData[] _vertices;
   FaceData[] _faces;

   static final int JMC_SIGNATURE_FILE_FIXED_32 = 0x66783332;       // fx32
   static final int JMC_SIG32_COMMON   = 0x4A4D0000;                // JM
   static final int JMC_SIG32_IMAGES   = JMC_SIG32_COMMON | 0x6900; //   i*
   static final int JMC_SIG32_TEXTURES = JMC_SIG32_COMMON | 0x7400; //   t*
   static final int JMC_SIG32_VERTICES = JMC_SIG32_COMMON | 0x7600; //   v*
   static final int JMC_SIG32_FACES    = JMC_SIG32_COMMON | 0x6100; //   a*
   static final int JMC_SIG32_END      = JMC_SIG32_COMMON | 0x6E69; //   fi
   static final int JMC_FACEFLAG_DOUBLESIDED = 0x80;
}

