package com.randomsoftware.common.meshlib;

import java.util.*;
import danger.app.*;
import danger.ui.*;
import danger.util.*;
import danger.audio.*;

public class Scene
{
   public Mesh firstMesh = null;
   public Mesh lastMesh = null;

   public Actor firstActor = null;
   public Actor lastActor = null;

   public Camera firstCamera = null;
   public Camera lastCamera = null;

   public Light firstLight = null;
   public Light lastLight = null;

   public Trigger firstTrigger = null;
   public Trigger lastTrigger = null;

   public Scene()
   {
   }

   public void destroy()
   {
      while (firstMesh != null)
         firstMesh.destroy();

      while (firstActor != null)
         firstActor.destroy();

      while (firstLight != null)
         firstLight.destroy();
   }

   public void update()
   {
      for (Actor actor = firstActor; actor != null; actor = actor.nextActorInScene)
         actor.update();
      for (Mesh mesh = firstMesh; mesh != null; mesh = mesh.nextMeshInScene)
         mesh.update();
   }

   public void invalidate()
   {
      _camLastFrame = null;
   }

   public void paint (Pen pp, Camera cam)
   {
      boolean fAllInvalid = (_camLastFrame != cam);
      _camLastFrame = cam;
      _fRenderingNow = true;

      int xBias = (cam.pos.right - cam.pos.left) / 2;
      int yBias = (cam.pos.bottom - cam.pos.top) / 2;

      // What is the transformation matrix from world space to camera space?

      FixedMatrix mxWorldToCamera = cam._mxMultiple;

      // Update all actors' locations in the world space.

      for (Actor actor = firstActor; actor != null; actor = actor.nextActorInScene)
      {
         if (fAllInvalid || actor.isChanged)
            actor.recalculate();
      }

      // Ensure our triggers are properly translated

      if (firstTrigger != null)
      {
         for (Trigger trigger = firstTrigger; trigger != null; trigger = trigger.nextTriggerInScene)
         {
            trigger.update();

            if (fAllInvalid || trigger._isChanged)
            {
               trigger.recalculate (mxWorldToCamera);
               trigger._dxScreen = xBias + (int)(trigger._vPosInCamera.dx >> FixedMath.SHIFT);
               trigger._dyScreen = yBias - (int)(trigger._vPosInCamera.dy >> FixedMath.SHIFT);
            }
         }
      }

      // Update all meshes.  If we have to recalculate a mesh,
      // we'll also tell it what we want to rebuild.
      // Here the mesh walks through its vertices and recalculates
      // each one's position in both world and camera space.  (We
      // need world-space data for lighting calculations, and camera-
      // space data so we can draw the faces.)

      boolean fRenderLights = (cam.lighting != Camera.LIGHTING_NONE) && (firstLight != null);
      boolean fCullBackface = ((cam.cull & Camera.CULL_BACKFACE) != 0);
      int flags = Mesh.MESH_NEED_CAMERA_VERTEXPOS;

      if (fRenderLights)
         flags |= Mesh.MESH_NEED_WORLD_FACENORMALS;
      if (fCullBackface)
         flags |= Mesh.MESH_NEED_CAMERA_FACENORMALS;

      for (Mesh mesh = firstMesh; mesh != null; mesh = mesh.nextMeshInScene)
      {
         int localFlags = flags;
         if (mesh.isUnlit)
            localFlags &= ~Mesh.MESH_NEED_WORLD_FACENORMALS;
         if (mesh.isChanged)
            mesh.recalculate (cam, mxWorldToCamera, localFlags);
         else if (fAllInvalid || mesh._fNeedsCameraPos)
            mesh.recalculate (cam, mxWorldToCamera, localFlags & (Mesh.MESH_NEED_CAMERA_VERTEXPOS | Mesh.MESH_NEED_CAMERA_FACENORMALS));
      }

      // We now know the camera-space normal vectors for all the faces of all
      // meshes. We'll use that for standard backface culling, and we'll decide
      // which vertices are unused because their faces were culled.

      for (Mesh mesh = firstMesh; mesh != null; mesh = mesh.nextMeshInScene)
      {
         if (!mesh._fWillRender)
            continue;

         if (fCullBackface)
         {
            for (Vertex vertex = mesh.firstVertex; vertex != null; vertex = vertex.nextVertexInMesh)
            {
               vertex._fWillRender = false;
            }

            for (Face face = mesh.firstFace; face != null; face = face.nextFaceInMesh)
            {
               if (face.isDoubleSided)
                  face._fWillRender = true;
               else if (face._vNormalInCamera.dz > 0)
                  face._fWillRender = true;
               else
                  face._fWillRender = false;

               if (face._fWillRender)
               {
                  face.vertex1._fWillRender = true;
                  face.vertex2._fWillRender = true;
                  face.vertex3._fWillRender = true;
               }
            }
         }
         else
         {
            for (Vertex vertex = mesh.firstVertex; vertex != null; vertex = vertex.nextVertexInMesh)
            {
               vertex._fWillRender = true;
            }
         }
      }

      // Sort all the faces in the scene: we need to draw the furthest
      // away, first.

      _facesort.update (this, !fCullBackface);

      // Transform into screen coordinates.  The camera's viewing matrix
      // will do perspective or orthographic projection for us.

      for (Mesh mesh = firstMesh; mesh != null; mesh = mesh.nextMeshInScene)
      {
         if (!mesh._fWillRender)
            continue;
         for (Vertex vertex = mesh.firstVertex; vertex != null; vertex = vertex.nextVertexInMesh)
         {
            if (!vertex._fWillRender)
               continue;

            if (vertex._vPosInCamera.dz >= 0-cam._dNear)
            {
               vertex._fWillRender = false;
               continue;
            }

            vertex._dxScreen = xBias + (int)(vertex._vPosInCamera.dx >> FixedMath.SHIFT);
            vertex._dyScreen = yBias - (int)(vertex._vPosInCamera.dy >> FixedMath.SHIFT);
         }
      }

      // If we've moved a mesh, we need to relight lots of faces

      if (cam.lighting != Camera.LIGHTING_NONE)
      {
         for (Light light = firstLight; light != null; light = light.nextLightInScene)
         {
            if (light._fNeedsRelight)
            {
               light.clearLights();

               for (Mesh mesh = firstMesh; mesh != null; mesh = mesh.nextMeshInScene)
               {
                  // Skip for now any meshes that need full relights;
                  // we'll do them later
                  if (!mesh._fWillRender)
                     continue;
                  if (mesh.isUnlit)
                     continue;
                  if (!mesh._fNeedsRelight)
                     light.relightMeshFaces (mesh, cam);
               }

               light._fNeedsRelight = false;
            }
         }

         for (Mesh mesh = firstMesh; mesh != null; mesh = mesh.nextMeshInScene)
         {
            if (!mesh._fWillRender)
               continue;
            if (mesh._fNeedsRelight && !mesh.isUnlit)
            {
               mesh.clearLights();

               for (Light light = firstLight; light != null; light = light.nextLightInScene)
                  light.relightMeshFaces (mesh, cam);

               mesh._fNeedsRelight = false;
            }
         }
      }

      // Now draw the faces themselves.  We've got two special-purpose render
      // loops up front, to fast-path the common rendering cases.

      if (cam.quality == Camera.QUALITY_POINT)
      {
         pp.setColor (Color.WHITE);

         for (Mesh mesh = firstMesh; mesh != null; mesh = mesh.nextMeshInScene)
         {
            if (!mesh._fWillRender)
               continue;

            for (Vertex vertex = mesh.firstVertex; vertex != null; vertex = vertex.nextVertexInMesh)
            {
               if (!vertex._fWillRender)
                  continue;
               pp.drawPoint (vertex._dxScreen, vertex._dyScreen);
            }
         }
      }
      else if (cam.quality == Camera.QUALITY_LINE)
      {
         pp.setColor (Color.WHITE);

         for (int ii = 0; ii < _facesort.sortedFacesInUse; ++ii)
         {
            if ((countTriggers != 0) && (_facesort.sortedFaces[ii] instanceof Trigger))
            {
               Trigger trigger = (Trigger)_facesort.sortedFaces[ii];
               trigger.paint (pp, trigger._dxScreen, trigger._dyScreen);
            }
            else
            {
               Face face = (Face)_facesort.sortedFaces[ii];
               if ((!face.vertex1._fWillRender) || (!face.vertex2._fWillRender) || (!face.vertex3._fWillRender))
                  continue;
               pp.drawLine (face.vertex1._dxScreen, face.vertex1._dyScreen, face.vertex2._dxScreen, face.vertex2._dyScreen);
               pp.drawLine (face.vertex2._dxScreen, face.vertex2._dyScreen, face.vertex3._dxScreen, face.vertex3._dyScreen);
               pp.drawLine (face.vertex3._dxScreen, face.vertex3._dyScreen, face.vertex1._dxScreen, face.vertex1._dyScreen);
            }
         }
      }
      else // (cam.quality == Camera.QUALITY_SOLID || cam.quality == Camera.QUALITY_SOLID_OUTLINED || cam.quality == Camera.QUALITY_TEXTURED || cam.quality == Camera.QUALITY_TEXTURED_OUTLINED)
      {
         for (int ii = 0; ii < _facesort.sortedFacesInUse; ++ii)
         {
            if ((countTriggers != 0) && (_facesort.sortedFaces[ii] instanceof Trigger))
            {
               Trigger trigger = (Trigger)_facesort.sortedFaces[ii];
               trigger.paint (pp, trigger._dxScreen, trigger._dyScreen);
            }
            else
            {
               Face face = (Face)_facesort.sortedFaces[ii];
               if ((!face.vertex1._fWillRender) || (!face.vertex2._fWillRender) || (!face.vertex3._fWillRender))
                  continue;

               if (face.isTextured() && (cam.quality == Camera.QUALITY_TEXTURED || cam.quality == Camera.QUALITY_TEXTURED_OUTLINED))
               {
                  // The Danger poly-filler requires clockwise rendering, but
                  // our visible triangles are always counter-clockwise on the
                  // screen so we have to flip a coordinate pair when drawing.
                  // Note that our texture coordinates are per-face, and have
                  // already been likewise flipped anticipating this step.
                  _axPoly[0] = face.vertex1._dxScreen;
                  _ayPoly[0] = face.vertex1._dyScreen;
                  _axPoly[1] = face.vertex3._dxScreen;
                  _ayPoly[1] = face.vertex3._dyScreen;
                  _axPoly[2] = face.vertex2._dxScreen;
                  _ayPoly[2] = face.vertex2._dyScreen;
                  face.paintTextured (pp, _axPoly, _ayPoly);
               }
               else
               {
                  _axPoly[0] = face.vertex1._dxScreen;
                  _ayPoly[0] = face.vertex1._dyScreen;
                  _axPoly[1] = face.vertex2._dxScreen;
                  _ayPoly[1] = face.vertex2._dyScreen;
                  _axPoly[2] = face.vertex3._dxScreen;
                  _ayPoly[2] = face.vertex3._dyScreen;

                  if (fRenderLights && !face._mesh.isUnlit)
                     face.texture.getColor (cam, face, Texture.GETCOLOR_FOG | Texture.GETCOLOR_LIGHTING, clr);
                  else
                     face.texture.getColor (cam, face, Texture.GETCOLOR_FOG, clr);
                  pp.setColor (clr);
                  face.paintSolid (pp, _axPoly, _ayPoly);
               }

               if ((cam.quality == Camera.QUALITY_SOLID_OUTLINED) || (cam.quality == Camera.QUALITY_TEXTURED_OUTLINED))
               {
                  pp.setColor (Color.BLACK);
                  pp.drawPoly (_axPoly, _ayPoly);
               }
            }
         }
      }

      _fRenderingNow = false;
   }

   Camera _camLastFrame = null;
   FaceSort _facesort = new FaceSort();
   FaceLightFactory _faceLightFactory = new FaceLightFactory();
   boolean _fRenderingNow = false;
   static int _axPoly[] = new int[3];
   static int _ayPoly[] = new int[3];
   static Color clr = new Color();
   public int countTriggers = 0;

   static final long MAX_CAMERA_DEPTH = FixedMath.PREC / 100;
}

