package com.randomsoftware.common.terrain;
import com.randomsoftware.common.animation.*;
import com.randomsoftware.common.meshlib.*;

import java.util.*;
import danger.app.*;
import danger.ui.*;
import danger.util.*;
import danger.audio.*;

public class GroundMesh extends Mesh
{
   public GroundMesh (Scene sc, HeightMap heightMap, TerrainMap terrainMap, GroundViewer viewer, TerrainConfig config)
   {
      super (sc);

      _scene = sc;
      _heightMap = heightMap;
      _terrainMap = terrainMap;
      _viewer = viewer;
      _config = config;
      _config.calculateDerivedValues();
      _grid = new GroundMeshGrid (this);

      _orientationLast = ~0;
      _xLast = ~0;
      _zLast = ~0;
   }

   public void destroy()
   {
      _grid.destroy();
      super.destroy();
   }

   public void update()
   {
      int flags = 0;

      if (_viewer.orientation != _orientationLast)
      {
         flags = MESH_INVALIDATE_CAMERA;
         _orientationLast = _viewer.orientation;
      }

      if ( ((_viewer.source.dx>>_config.SHIFT_UPDATE_TEST) != (_xLast>>_config.SHIFT_UPDATE_TEST)) ||
           ((_viewer.source.dz>>_config.SHIFT_UPDATE_TEST) != (_zLast>>_config.SHIFT_UPDATE_TEST)) ||
           (flags != 0) )
      {
         _xLast = _viewer.source.dx;
         _zLast = _viewer.source.dz;

         if (reposition (_xLast, _zLast))
         {
            flags |= MESH_INVALIDATE_DEFORM;
         }
      }

      if (flags != 0)
      {
         invalidate (flags);
      }
   }

   boolean reposition (long xStand, long zStand)
   {
      boolean adjusted = false;

      long NEAR_FOV = FixedMath.HEXIAN_90;
      long NEAR_RANGE_UNSHIFTED = 5;
      long FAR_RANGE_UNSHIFTED = ((_config.SCALE * (_config.MAJOR_GROUP_COUNT-2)) - 1) >> FixedMath.SHIFT;

      // Find the rough rectangle which we need to fill with triangles

      long xLeft = xStand;
      long xRight = xStand;
      long zTop = zStand;
      long zBottom = zStand;

      long xx,zz;
      xx = xStand + FixedMath.cosine(_viewer.orientation+NEAR_FOV/2) * NEAR_RANGE_UNSHIFTED;
      zz = zStand + FixedMath.sine(_viewer.orientation+NEAR_FOV/2) * NEAR_RANGE_UNSHIFTED;
      if (xx < xLeft) xLeft = xx;
      if (xx > xRight) xRight = xx;
      if (zz < zTop) zTop = zz;
      if (zz > zBottom) zBottom = zz;

      xx = xStand + FixedMath.cosine(_viewer.orientation-NEAR_FOV/2) * FAR_RANGE_UNSHIFTED;
      zz = zStand + FixedMath.sine(_viewer.orientation-NEAR_FOV/2) * FAR_RANGE_UNSHIFTED;
      if (xx < xLeft) xLeft = xx;
      if (xx > xRight) xRight = xx;
      if (zz < zTop) zTop = zz;
      if (zz > zBottom) zBottom = zz;

      xx = xStand + FixedMath.cosine(_viewer.orientation) * FAR_RANGE_UNSHIFTED;
      zz = zStand + FixedMath.sine(_viewer.orientation) * FAR_RANGE_UNSHIFTED;
      if (xx < xLeft) xLeft = xx;
      if (xx > xRight) xRight = xx;
      if (zz < zTop) zTop = zz;
      if (zz > zBottom) zBottom = zz;

      xx = xStand + FixedMath.cosine(_viewer.orientation+_viewer.fov/2) * FAR_RANGE_UNSHIFTED;
      zz = zStand + FixedMath.sine(_viewer.orientation+_viewer.fov/2) * FAR_RANGE_UNSHIFTED;
      if (xx < xLeft) xLeft = xx;
      if (xx > xRight) xRight = xx;
      if (zz < zTop) zTop = zz;
      if (zz > zBottom) zBottom = zz;

      xx = xStand + FixedMath.cosine(_viewer.orientation-_viewer.fov/2) * FAR_RANGE_UNSHIFTED;
      zz = zStand + FixedMath.sine(_viewer.orientation-_viewer.fov/2) * FAR_RANGE_UNSHIFTED;
      if (xx < xLeft) xLeft = xx;
      if (xx > xRight) xRight = xx;
      if (zz < zTop) zTop = zz;
      if (zz > zBottom) zBottom = zz;

      // Expand that area to land on MAX_SCALE boundaries.

      xLeft = roundDown(xLeft, _config.SCALE);
      zTop = roundDown(zTop, _config.SCALE);
      xRight = roundUp(xRight, _config.SCALE);
      zBottom = roundUp(zBottom, _config.SCALE);

      // Just to be safe, we'll ensure we don't exceed our maximum size.
      // Calculations are such that we could put an assert here, but I
      // like being defensive.

      int cx = (int)( ((xRight+1) - xLeft) / _config.SCALE );
      if (cx > _config.MAJOR_GROUP_COUNT)
      {
         cx = _config.MAJOR_GROUP_COUNT;
         xRight = xLeft + cx * _config.SCALE;
      }

      int cz = (int)( ((zBottom+1) - zTop) / _config.SCALE );
      if (cz > _config.MAJOR_GROUP_COUNT)
      {
         cz = _config.MAJOR_GROUP_COUNT;
         zBottom = zTop + cz * _config.SCALE;
      }

      // The overall mesh itself must be centered somewhere in front of
      // the camera, or it will be culled immediately by the rendering engine.

      long xMeshCenter = (xLeft + xRight) / 2;
      long zMeshCenter = (zTop + zBottom) / 2;

      // We've found the greater area that we're planning on filling in.
      // Odds are that's the same big area as last time (the boundaries
      // are on nearby SCALE divisions, which are pretty wide--so you
      // have to turn or travel a bit before they change).  If it's not
      // the same area as last time, though, blow away everything we had
      // earlier.

      if ((_grid._xLeft != xLeft) || (_grid._xRight != xRight) || (_grid._zTop != zTop) || (_grid._zBottom != zBottom))
      {
         setLocation (xMeshCenter, 0, zMeshCenter);
         _grid.reset (xMeshCenter, zMeshCenter, xLeft, zTop, xRight, zBottom);
         adjusted = true;
      }

      // Now the easy part: just tell the grid to get populated.

      if (_grid.update (_viewer.target.dx, _viewer.target.dz))
         adjusted = true;

      return adjusted;
   }

   static final long roundDown (long vv, long gran)
   {
      if (vv < 0)
         return 0 - roundUp (0-vv, gran);
      else
         return vv - (vv % gran);
   }

   static final long roundUp (long vv, long gran)
   {
      if (vv < 0)
         return 0 - roundDown (0-vv, gran);
      else
         return (vv+gran) - ((vv+gran-1) % gran);
   }

   Scene _scene;
   GroundViewer _viewer;
   HeightMap _heightMap;
   TerrainMap _terrainMap;
   GroundMeshGrid _grid;
   TerrainConfig _config;

   long _xLast;
   long _zLast;
   long _orientationLast;
}

